from dataclasses import dataclass
from typing import Optional, List, Type
from bloqade.builder.base import Builder


@dataclass
class BuilderNode:
    node: Builder
    next: Optional["BuilderNode"] = None

    def __repr__(self) -> str:
        return repr(self.node)


@dataclass
class BuilderStream:
    """Represents a stream of builder nodes."""

    head: BuilderNode
    curr: Optional[BuilderNode] = None

    def copy(self) -> "BuilderStream":
        return BuilderStream(head=self.head, curr=self.curr)

    def read(self) -> Optional[BuilderNode]:
        if self.curr is None:
            return None

        node = self.curr
        self.curr = self.curr.next
        return node

    def read_next(self, builder_types: List[type[Builder]]) -> Optional[BuilderNode]:
        node = self.read()
        while node is not None:
            if type(node.node) in builder_types:
                return node
            node = self.read()
        return None

    def eat(
        self, types: List[Type[Builder]], skips: Optional[List[Type[Builder]]] = None
    ) -> BuilderNode:
        """Scan the stream until a node of type in `types` or `skips` is found.

        Args:
            types (List[Type[Builder]]): List of types to move the stream pointer to
            skips (List[Type[Builder]] | None, optional): List of types to end the
            stream scan

        Returns:
            BuilderNode: The beginning of the stream which matches a type in `types`.
        """
        head = self.read_next(types)
        curr = head
        while curr is not None:
            if type(curr.node) not in types:
                if skips and type(curr.node) not in skips:
                    break
            curr = curr.next
        self.curr = curr
        return head

    def __iter__(self):
        return self

    def __next__(self):
        node = self.read()
        if node is None:
            raise StopIteration
        return node

    @staticmethod
    def build_nodes(ast: Builder) -> "BuilderNode":
        curr = ast
        node = None
        while curr is not None:
            next = curr
            curr = curr.__parent__
            node = BuilderNode(next, node)

        return node

    @staticmethod
    def create(builder: Builder) -> "BuilderStream":
        head = BuilderStream.build_nodes(builder)
        return BuilderStream(head=head, curr=head)



from .base import AtomArrangement, ParallelRegister, LocationInfo
from .bravais import (
    BoundedBravais,
    Chain,
    Square,
    Rectangular,
    Honeycomb,
    Triangular,
    Lieb,
    Kagome,
)
from .list import ListOfLocations, start

__all__ = [
    "start",
    "AtomArrangement",
    "Chain",
    "Square",
    "Rectangular",
    "Honeycomb",
    "Triangular",
    "Lieb",
    "Kagome",
    "BoundedBravais",
    "ListOfLocations",
    "ParallelRegister",
    "LocationInfo",
]



# Ported from the Julia language AbstractTrees.jl implementation: https://github.com/JuliaCollections/AbstractTrees.jl/blob/master/src/printing.jl

from pydantic.dataclasses import dataclass
import sys
import bloqade.ir as ir

# The maximum depth of the tree to print.
MAX_TREE_DEPTH = 10


unicode_enabled = sys.stdout.encoding.lower().startswith("utf")
color_enabled = ("ipykernel" in sys.modules) or sys.stdout.isatty


class KeyValuePair:
    def __init__(self, key, value):
        self.key = key
        self.value = value

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, KeyValuePair):
            return False

        return self.key == __value.key and self.value == __value.value

    def children(self):
        return {"key": self.key, "value": self.value}

    def print_node(self):
        return "KeyValuePair"


# charset object, extracts away charset
@dataclass
class UnicodeCharSet:
    mid = "├"
    terminator = "└"
    skip = "│"
    dash = "─"
    trunc = "⋮"
    pair = "⇒ "


@dataclass
class ASCIICharSet:
    mid = "+"
    terminator = "\\"
    skip = "|"
    dash = "--"
    trunc = "..."
    pair = "=> "


@dataclass
class State:
    depth: int = 0
    prefix: str = ""
    last: bool = False


@dataclass
class ColorCode:
    CEND = "\33[0m"
    CBOLD = "\33[1m"
    CITALIC = "\33[3m"
    CURL = "\33[4m"
    CBLINK = "\33[5m"
    CBLINK2 = "\33[6m"
    CSELECTED = "\33[7m"

    CBLACK = "\33[30m"
    CRED = "\33[31m"
    CGREEN = "\33[32m"
    CYELLOW = "\33[33m"
    CBLUE = "\33[34m"
    CVIOLET = "\33[35m"
    CBEIGE = "\33[36m"
    CWHITE = "\33[37m"

    CBLACKBG = "\33[40m"
    CREDBG = "\33[41m"
    CGREENBG = "\33[42m"
    CYELLOWBG = "\33[43m"
    CBLUEBG = "\33[44m"
    CVIOLETBG = "\33[45m"
    CBEIGEBG = "\33[46m"
    CWHITEBG = "\33[47m"

    CGREY = "\33[90m"
    CRED2 = "\33[91m"
    CGREEN2 = "\33[92m"
    CYELLOW2 = "\33[93m"
    CBLUE2 = "\33[94m"
    CVIOLET2 = "\33[95m"
    CBEIGE2 = "\33[96m"
    CWHITE2 = "\33[97m"

    CGREYBG = "\33[100m"
    CREDBG2 = "\33[101m"
    CGREENBG2 = "\33[102m"
    CYELLOWBG2 = "\33[103m"
    CBLUEBG2 = "\33[104m"
    CVIOLETBG2 = "\33[105m"
    CBEIGEBG2 = "\33[106m"
    CWHITEBG2 = "\33[107m"

    def set_color(self, color, str):
        return color + str + self.CEND


# all nodes should implement:
# children method  -> either list or Dict
# print_node method -> should just return str
# print_annotation ->  should just return str


# p is needed for IPython Pretty printer or
# the following helper wrap, which collect the tree,
# and spit out the tree as raw string
class MockPrinter:
    def __init__(self):
        self.str_out = ""

    def text(self, segs: str):
        self.str_out += segs

    def get_value(self):
        return self.str_out

    # @property
    # def cycle_byterm(self):
    #    return int(os.get_terminal_size().columns / 10)


class Printer:
    def __init__(self, p=None, max_tree_depth=None):
        self.charset = UnicodeCharSet() if unicode_enabled else ASCIICharSet()
        self.colorcode = ColorCode()
        self.state = State()
        if p is None:
            self.p = MockPrinter()
        else:
            self.p = p
        self.max_tree_depth = (
            MAX_TREE_DEPTH if max_tree_depth is None else max_tree_depth
        )

    def should_print_annotation(self, children):
        if isinstance(
            children, (list, tuple)
        ):  # or generator, not sure of equivalence in Python
            return False
        elif isinstance(children, dict):
            return True

    def get_value(self):
        return self.p.get_value()

    def print(self, node, cycle=None):
        if cycle is None or isinstance(cycle, bool):
            cycle = MAX_TREE_DEPTH

        # list of children
        children = node.children()
        node_str = node.print_node()

        if color_enabled:
            if isinstance(node, ir.Variable):
                node_str = self.colorcode.set_color(self.colorcode.CBLUE2, node_str)

        for i, line in enumerate(node_str.split("\n")):
            i != 0 and print(self.state.prefix)
            self.p.text(line)
            if not (self.state.last and len(children) == 0):
                self.p.text("\n")

        if self.state.depth > min(cycle, self.max_tree_depth):
            # need to set this variable dynamically
            self.p.text(self.charset.trunc)
            self.p.text("\n")
            return

        this_print_annotation = self.should_print_annotation(children)

        if this_print_annotation:
            children = list(children.items())
        else:
            children = list(children)

        while not len(children) == 0:
            child_prefix = self.state.prefix
            if this_print_annotation:
                annotation, child = children.pop(0)
            else:
                child = children.pop(0)
                annotation = None

            self.p.text(self.state.prefix)

            if len(children) == 0:
                self.p.text(self.charset.terminator)
                child_prefix += " " * (
                    len(self.charset.skip) + len(self.charset.dash) + 1
                )

                if self.state.depth > 0 and self.state.last:
                    is_last_leaf_child = True
                elif self.state.depth == 0:
                    is_last_leaf_child = True
                else:
                    is_last_leaf_child = False
            else:
                self.p.text(self.charset.mid)
                child_prefix += self.charset.skip + " " * (len(self.charset.dash) + 1)
                is_last_leaf_child = False

            self.p.text(self.charset.dash + " ")

            if this_print_annotation and annotation is not None:
                self.p.text(annotation)
                self.p.text("\n")
                self.p.text(child_prefix + self.charset.pair)
                child_prefix += " " * (len(self.charset.pair))

            self.state.depth += 1
            parent_last = self.state.last
            self.state.last = is_last_leaf_child
            parent_prefix = self.state.prefix
            self.state.prefix = child_prefix

            if not isinstance(child, str):
                self.print(child, cycle)
            else:
                self.p.text(child)
                self.p.text("\n")

            self.state.depth -= 1
            self.state.prefix = parent_prefix
            self.state.last = parent_last



from bloqade.ir import Field, Uniform, Linear, Pulse, NamedPulse, detuning, rabi
from bloqade.ir import Interval
from bloqade import cast
import pytest
from io import StringIO
from IPython.lib.pretty import PrettyPrinter as PP
from bloqade.ir.control.pulse import (
    FieldName,
    RabiFrequencyAmplitude,
    RabiFrequencyPhase,
)
import bloqade.ir.tree_print as trp

trp.color_enabled = False


def test_fieldname_base():
    x = FieldName()

    assert x.children() == []


## from top to bottom test each hierarchy classes:
# rabi (RabiRouter) |- RabiAmplitude
#                   |- RabiPhase
def test_rabi_router():
    R = rabi

    assert R.print_node() == "RabiRouter"
    assert R.children() == {
        "Amplitude": RabiFrequencyAmplitude(),
        "Phase": RabiFrequencyPhase(),
    }

    mystdout = StringIO()
    p = PP(mystdout)

    R._repr_pretty_(p, 10)

    assert (
        mystdout.getvalue()
        == "RabiRouter\n"
        + "├─ Amplitude\n"
        + "│  ⇒ RabiFrequencyAmplitude\n"
        + "└─ Phase\n"
        + "   ⇒ RabiFrequencyPhase"
    )


def test_rabi_amp_phase():
    rf = RabiFrequencyAmplitude()

    assert rf.print_node() == "RabiFrequencyAmplitude"

    mystdout = StringIO()
    p = PP(mystdout)

    rf._repr_pretty_(p, 10)

    assert mystdout.getvalue() == "RabiFrequencyAmplitude\n"

    rp = RabiFrequencyPhase()
    assert rp.print_node() == "RabiFrequencyPhase"

    mystdout = StringIO()
    p = PP(mystdout)

    rp._repr_pretty_(p, 10)

    assert mystdout.getvalue() == "RabiFrequencyPhase\n"


def test_detune():
    D = detuning

    assert D.print_node() == "Detuning"

    mystdout = StringIO()
    p = PP(mystdout)

    D._repr_pretty_(p, 10)

    assert mystdout.getvalue() == "Detuning\n"


def test_pulse():
    f = Field({Uniform: Linear(start=1.0, stop="x", duration=3.0)})

    ## make pulse, invalid field:
    with pytest.raises(TypeError):
        Pulse({Uniform: 1.0})

    ## make pulse:
    ps1 = Pulse({detuning: f})

    assert ps1.print_node() == "Pulse"
    assert ps1.children() == {"Detuning": f}

    mystdout = StringIO()
    p = PP(mystdout)

    ps1._repr_pretty_(p, 10)

    assert (
        mystdout.getvalue() == "Pulse\n"
        "└─ Detuning\n"
        "   ⇒ Field\n"
        "     └─ Drive\n"
        "        ├─ modulation\n"
        "        │  ⇒ UniformModulation\n"
        "        └─ waveform\n"
        "           ⇒ Linear\n"
        "             ├─ start\n"
        "             │  ⇒ Literal: 1.0\n"
        "             ├─ stop\n"
        "             │  ⇒ Variable: x\n"
        "             └─ duration\n"
        "                ⇒ Literal: 3.0"
    )


def test_named_pulse():
    f = Field({Uniform: Linear(start=1.0, stop="x", duration=3.0)})
    ps1 = Pulse({detuning: f})
    ps = NamedPulse("qq", ps1)

    assert ps.children() == {"Name": "qq", "Pulse": ps1}

    assert ps.print_node() == "NamedPulse"

    mystdout = StringIO()
    p = PP(mystdout)

    ps._repr_pretty_(p, 10)

    assert (
        mystdout.getvalue() == "NamedPulse\n"
        "├─ Name\n"
        "│  ⇒ qq\n"
        "└─ Pulse\n"
        "   ⇒ Pulse\n"
        "     └─ Detuning\n"
        "        ⇒ Field\n"
        "          └─ Drive\n"
        "             ├─ modulation\n"
        "             │  ⇒ UniformModulation\n"
        "             └─ waveform\n"
        "                ⇒ Linear\n"
        "                  ├─ start\n"
        "                  │  ⇒ Literal: 1.0\n"
        "                  ├─ stop\n"
        "                  │  ⇒ Variable: x\n"
        "                  └─ duration\n"
        "                     ⇒ Literal: 3.0"
    )


def test_slice_pulse():
    f = Field({Uniform: Linear(start=1.0, stop="x", duration=3.0)})
    ps1 = Pulse({detuning: f})
    itvl = Interval(cast(0), cast(1.5))

    ## invoke slice
    ps = ps1.slice(itvl)

    assert ps.print_node() == "Slice"
    assert ps.children() == {"Pulse": ps1, "Interval": itvl}

    mystdout = StringIO()
    p = PP(mystdout)

    ps._repr_pretty_(p, 10)

    assert (
        mystdout.getvalue() == "Slice\n"
        "├─ Pulse\n"
        "│  ⇒ Pulse\n"
        "│    └─ Detuning\n"
        "│       ⇒ Field\n"
        "│         └─ Drive\n"
        "│            ├─ modulation\n"
        "│            │  ⇒ UniformModulation\n"
        "│            └─ waveform\n"
        "│               ⇒ Linear\n"
        "│                 ├─ start\n"
        "│                 │  ⇒ Literal: 1.0\n"
        "│                 ├─ stop\n"
        "│                 │  ⇒ Variable: x\n"
        "│                 └─ duration\n"
        "│                    ⇒ Literal: 3.0\n"
        "└─ Interval\n"
        "   ⇒ Interval\n"
        "     ├─ start\n"
        "     │  ⇒ Literal: 0\n"
        "     └─ stop\n"
        "        ⇒ Literal: 1.5"
    )


def test_append_pulse():
    f = Field({Uniform: Linear(start=1.0, stop="x", duration=3.0)})
    ps1 = Pulse({detuning: f})

    # invoke append:
    ps = ps1.append(ps1)

    assert ps.children() == [ps1, ps1]

    mystdout = StringIO()
    p = PP(mystdout)

    ps._repr_pretty_(p, 10)
    print(repr(mystdout.getvalue()))
    assert (
        mystdout.getvalue() == "Append\n"
        "├─ Pulse\n"
        "│  └─ Detuning\n"
        "│     ⇒ Field\n"
        "│       └─ Drive\n"
        "│          ├─ modulation\n"
        "│          │  ⇒ UniformModulation\n"
        "│          └─ waveform\n"
        "│             ⇒ Linear\n"
        "│               ├─ start\n"
        "│               │  ⇒ Literal: 1.0\n"
        "│               ├─ stop\n"
        "│               │  ⇒ Variable: x\n"
        "│               └─ duration\n"
        "│                  ⇒ Literal: 3.0\n"
        "└─ Pulse\n"
        "   └─ Detuning\n"
        "      ⇒ Field\n"
        "        └─ Drive\n"
        "           ├─ modulation\n"
        "           │  ⇒ UniformModulation\n"
        "           └─ waveform\n"
        "              ⇒ Linear\n"
        "                ├─ start\n"
        "                │  ⇒ Literal: 1.0\n"
        "                ├─ stop\n"
        "                │  ⇒ Variable: x\n"
        "                └─ duration\n"
        "                   ⇒ Literal: 3.0"
    )


# print(Pulse({detuning: f}))
# print(Pulse({rabi.amplitude: f}))
# print(Pulse({rabi.phase: f}))
# print(Pulse({detuning: {Uniform: Linear(start=1.0, stop="x", duration=3.0)}}))



from bloqade.builder.base import Builder


class BloqadeService(Builder):
    @property
    def bloqade(self):
        """
        Specify the Bloqade backend.

        - Possible Next Steps:
            - `...bloqade.python()`: target submission to the Bloqade python backend
            - `...bloqade.julia()`: (CURRENTLY NOT IMPLEMENTED!)target
                submission to the Bloqade.jl backend
        """
        return BloqadeDeviceRoute(self)


class BloqadeDeviceRoute(Builder):
    def python(self):
        """
        Specify the Bloqade Python backend.

        - Possible Next Steps:
            - `...python().run(shots)`:
                to submit to the python emulator and await results
        """
        return self.parse().bloqade.python()

    def julia(self, solver: str, nthreads: int = 1):
        raise NotImplementedError



from collections import OrderedDict
from pydantic.dataclasses import dataclass
import json

from bloqade.builder.typing import LiteralType
from bloqade.ir.routine.base import RoutineBase, __pydantic_dataclass_config__
from bloqade.submission.quera import QuEraBackend
from bloqade.submission.mock import MockBackend
from bloqade.submission.quera_api_client.load_config import load_config
from bloqade.task.batch import RemoteBatch
from bloqade.task.quera import QuEraTask

from beartype.typing import Tuple, Union, Optional
from beartype import beartype


@dataclass(frozen=True, config=__pydantic_dataclass_config__)
class QuEraServiceOptions(RoutineBase):
    @beartype
    def device(self, config_file: Optional[str], **api_config):
        if config_file is not None:
            with open(config_file, "r") as f:
                api_config = {**json.load(f), **api_config}

        backend = QuEraBackend(**api_config)

        return QuEraHardwareRoutine(self.source, self.circuit, self.params, backend)

    def aquila(self) -> "QuEraHardwareRoutine":
        backend = QuEraBackend(**load_config("Aquila"))
        return QuEraHardwareRoutine(self.source, self.circuit, self.params, backend)

    def cloud_mock(self) -> "QuEraHardwareRoutine":
        backend = QuEraBackend(**load_config("Mock"))
        return QuEraHardwareRoutine(self.source, self.circuit, self.params, backend)

    @beartype
    def mock(
        self, state_file: str = ".mock_state.txt", submission_error: bool = False
    ) -> "QuEraHardwareRoutine":
        backend = MockBackend(state_file=state_file, submission_error=submission_error)
        return QuEraHardwareRoutine(self.source, self.circuit, self.params, backend)


@dataclass(frozen=True, config=__pydantic_dataclass_config__)
class QuEraHardwareRoutine(RoutineBase):
    backend: Union[QuEraBackend, MockBackend]

    def _compile(
        self,
        shots: int,
        args: Tuple[LiteralType, ...] = (),
        name: Optional[str] = None,
    ) -> RemoteBatch:
        from bloqade.codegen.common.assign_variables import AssignAnalogCircuit

        from bloqade.ir.analysis.assignment_scan import AssignmentScan
        from bloqade.codegen.hardware.quera import AHSCodegen

        circuit, params = self.circuit, self.params
        capabilities = self.backend.get_capabilities()
        circuit = AssignAnalogCircuit(params.static_params).visit(circuit)

        tasks = OrderedDict()

        for task_number, batch_params in enumerate(params.batch_assignments(*args)):
            record_params = AssignmentScan(batch_params).emit(circuit)
            final_circuit = AssignAnalogCircuit(record_params).visit(circuit)
            result = AHSCodegen(shots, capabilities=capabilities).emit(final_circuit)
            task_ir = result.quera_task_ir.discretize(capabilities)
            metadata = {**params.static_params, **record_params}
            tasks[task_number] = QuEraTask(
                None, self.backend, task_ir, metadata, result.parallel_decoder
            )

        batch = RemoteBatch(source=self.source, tasks=tasks, name=name)

        return batch

    @beartype
    def run_async(
        self,
        shots: int,
        args: Tuple[LiteralType, ...] = (),
        name: Optional[str] = None,
        shuffle: bool = False,
        **kwargs,
    ) -> RemoteBatch:
        """
        Compile to a RemoteBatch, which contain
            QuEra backend specific tasks,
            and run_async through QuEra service.

        Args:
            shots (int): number of shots
            args (Tuple): additional arguments
            name (str): custom name of the batch
            shuffle (bool): shuffle the order of jobs

        Return:
            RemoteBatch

        """
        batch = self._compile(shots, args, name)
        batch._submit(shuffle, **kwargs)
        return batch

    @beartype
    def run(
        self,
        shots: int,
        args: Tuple[LiteralType, ...] = (),
        name: Optional[str] = None,
        shuffle: bool = False,
        **kwargs,
    ) -> RemoteBatch:
        batch = self.run_async(shots, args, name, shuffle, **kwargs)
        batch.pull()
        return batch

    @beartype
    def __call__(
        self,
        *args: LiteralType,
        shots: int = 1,
        name: Optional[str] = None,
        shuffle: bool = False,
        **kwargs,
    ) -> RemoteBatch:
        return self.run(shots, args, name, shuffle, **kwargs)



from functools import cached_property
from ..scalar import Scalar, cast
from ..tree_print import Printer
from .waveform import Waveform
from bloqade.visualization import get_field_figure
from pydantic.dataclasses import dataclass
from typing import Dict, List, Optional
from decimal import Decimal
from bloqade.visualization import display_ir
from bloqade.visualization import get_ir_figure


class FieldExpr:
    def __str__(self):
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)


__all__ = [
    "Field",
    "Location",
    "SpatialModulation",
    "Uniform",
    "RunTimeVector",
    "ScaledLocations",
]


@dataclass(frozen=True)
class Location(FieldExpr):
    value: int

    def __str__(self):
        return f"Location({str(self.value)})"

    def children(self):
        return []

    def print_node(self):
        return f"Location({str(self.value)})"


@dataclass
class SpatialModulation(FieldExpr):
    def __hash__(self) -> int:
        raise NotImplementedError

    def _get_data(self, **assignment):
        return {}

    def figure(self, **assignment):
        raise NotImplementedError


@dataclass
class UniformModulation(SpatialModulation):
    def __hash__(self) -> int:
        return hash(self.__class__)

    def print_node(self):
        return "UniformModulation"

    def children(self):
        return []

    def _get_data(self, **assignment):
        return ["uni"], ["all"]

    def figure(self, **assignment):
        return get_ir_figure(self, **assignment)

    def show(self, **assignment):
        display_ir(self, **assignment)


Uniform = UniformModulation()


@dataclass
class RunTimeVector(SpatialModulation):
    name: str

    def __hash__(self) -> int:
        return hash(self.name) ^ hash(self.__class__)

    def print_node(self):
        return "RunTimeVector"

    def children(self):
        return [self.name]

    def figure(self, **assginment):
        return get_ir_figure(self, **assginment)

    def _get_data(self, **assignment):
        return [self.name], ["vec"]

    def show(self, **assignment):
        display_ir(self, **assignment)


@dataclass
class AssignedRunTimeVector(SpatialModulation):
    name: Optional[str]  # name is optional for literal Lists
    value: List[Decimal]

    @cached_property
    def _hash_value(self):
        return hash(self.name) ^ hash(tuple(self.value)) ^ hash(self.__class__)

    def __hash__(self) -> int:
        return self._hash_value

    def print_node(self):
        return "AssgiendRunTimeVector"

    def children(self):
        return [self.name, *self.value]

    def figure(self, **assginment):
        return get_ir_figure(self, **assginment)

    def _get_data(self, **assignment):
        return [self.name], ["vec"]

    def show(self, **assignment):
        display_ir(self, **assignment)


@dataclass(init=False)
class ScaledLocations(SpatialModulation):
    value: Dict[Location, Scalar]

    def __init__(self, pairs):
        value = dict()
        for k, v in pairs.items():
            if isinstance(k, int):
                k = Location(k)
            elif isinstance(k, Location):
                pass
            else:
                raise ValueError(f"expected Location or int, got {k}")

            value[k] = cast(v)
        self.value = value

    @cached_property
    def _hash_value(self) -> int:
        return hash(frozenset(self.value.items())) ^ hash(self.__class__)

    def __hash__(self) -> int:
        return self._hash_value

    def _get_data(self, **assignments):
        names = []
        scls = []

        for loc, scl in self.value.items():
            names.append("loc[%d]" % (loc.value))
            scls.append(str(scl(**assignments)))

        return names, scls

    def print_node(self):
        return "ScaledLocations"

    def children(self):
        # can return list or dict
        # should return dict consisting of Location and Scalar
        annotated_children = {}
        for loc, scalar in self.value.items():
            annotated_children[loc.print_node()] = scalar

        return annotated_children

    def __getitem__(self, key):
        return self.value[key]

    def __setitem__(self, key, value):
        assert isinstance(key, Location)
        self.value[key] = cast(value)

    def __bool__(self):
        return bool(self.value)

    def figure(self, **assignments):
        return get_ir_figure(self, **assignments)

    def show(self, **assignment):
        display_ir(self, assignment)


@dataclass
class Drive:
    modulation: SpatialModulation
    waveform: Waveform

    def print_node(self):
        return "Drive"

    def children(self):
        return {"modulation": self.modulation, "waveform": self.waveform}


@dataclass(frozen=True)
class Field(FieldExpr):
    """Field node in the IR. Which contains collection(s) of
    [`Waveform`][bloqade.ir.control.waveform.Waveform]

    ```bnf
    <field> ::= ('field' <spatial modulation>  <padded waveform>)*
    ```
    """

    drives: Dict[SpatialModulation, Waveform]

    @cached_property
    def _hash_value(self):
        return hash(frozenset(self.drives.items())) ^ hash(self.__class__)

    def __hash__(self) -> int:
        return self._hash_value

    def canonicalize(self) -> "Field":
        """
        Canonicalize the Field by merging `ScaledLocation` nodes with the same waveform.
        """
        reversed_dirves = {}

        for sm, wf in self.drives.items():
            reversed_dirves[wf] = reversed_dirves.get(wf, []) + [sm]

        drives = {}

        for wf, sms in reversed_dirves.items():
            new_sm = [sm for sm in sms if not isinstance(sm, ScaledLocations)]
            scaled_locations_sm = [sm for sm in sms if isinstance(sm, ScaledLocations)]

            new_mask = {}

            for ele in scaled_locations_sm:
                for loc, scl in ele.value.items():
                    new_mask[loc] = new_mask.get(loc, 0) + cast(scl)

            if new_mask:
                new_sm += [ScaledLocations(new_mask)]

            for sm in new_sm:
                drives[sm] = wf

        return Field(drives)

    def add(self, other):
        if not isinstance(other, Field):
            raise ValueError(f"Cannot add Field and {other.__class__}")

        out = Field(dict(self.drives))

        for spatial_modulation, waveform in other.drives.items():
            if spatial_modulation in self.drives:
                out.drives[spatial_modulation] = (
                    out.drives[spatial_modulation] + waveform
                )
            else:
                out.drives[spatial_modulation] = waveform

        return out.canonicalize()

    def print_node(self):
        return "Field"

    def children(self):
        # return dict with annotations
        return [Drive(k, v) for k, v in self.drives.items()]

    def figure(self, **assignments):
        return get_field_figure(self, "Field", None, **assignments)

    def show(self, **assignments):
        """
        Interactive visualization of the Field

        Args:
            **assignments: assigning the instance value (literal) to the
                existing variables in the Field

        """
        display_ir(self, assignments)



from collections import OrderedDict
from decimal import Decimal
from bloqade.submission.braket import BraketBackend
from bloqade.submission.ir.braket import BraketTaskSpecification
from bloqade.submission.ir.parallel import ParallelDecoder
from bloqade.submission.ir.task_results import QuEraTaskResults
from bloqade.submission.ir.task_specification import QuEraTaskSpecification
from bloqade.submission.mock import MockBackend
from bloqade.submission.quera import QuEraBackend
from .braket import BraketTask
from .quera import QuEraTask
from .batch import RemoteBatch, LocalBatch
from typing import TextIO, Dict, Any, Union
from .braket_simulator import BraketEmulatorTask
import json


class BatchSerializer(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Decimal):
            return str(obj)
        elif isinstance(obj, QuEraBackend):
            return {
                "quera_backend": obj.dict(
                    exclude=set(["access_key", "secret_key", "session_token"])
                )
            }
        elif isinstance(obj, MockBackend):
            return {"mock_backend": obj.dict()}
        elif isinstance(obj, BraketBackend):
            return {"braket_backend": obj.dict()}
        elif isinstance(obj, QuEraTaskSpecification):
            return {"quera_task_specification": obj.dict(by_alias=True)}
        elif isinstance(obj, BraketTaskSpecification):
            return {"braket_task_specification": obj.dict()}
        elif isinstance(obj, ParallelDecoder):
            return {"parallel_decoder": obj.dict()}
        elif isinstance(obj, QuEraTaskResults):
            return {"task_result_ir": obj.dict()}
        elif isinstance(obj, BraketTask):
            return {
                "braket_task": {
                    "task_id": obj.task_id,
                    "backend": obj.backend,
                    "parallel_decoder": obj.parallel_decoder,
                    "task_result_ir": obj.task_result_ir,
                    "task_ir": obj.task_ir,
                    "metadata": obj.metadata,
                }
            }
        elif isinstance(obj, BraketEmulatorTask):
            return {
                "braket_emulator_task": {
                    "task_ir": obj.task_ir,
                    "metadata": obj.metadata,
                    "task_result_ir": obj.task_result_ir,
                }
            }
        elif isinstance(obj, QuEraTask):
            return {
                "quera_task": {
                    "backend": obj.backend,
                    "task_id": obj.task_id,
                    "parallel_decoder": obj.parallel_decoder,
                    "task_result_ir": obj.task_result_ir,
                    "task_ir": obj.task_ir,
                    "metadata": obj.metadata,
                }
            }

        elif isinstance(obj, LocalBatch):
            return {
                "local_batch": {
                    "source": None,
                    "name": obj.name,
                    "tasks": [(k, v) for k, v in obj.tasks.items()],
                }
            }
        elif isinstance(obj, RemoteBatch):
            return {
                "remote_batch": {
                    "source": None,
                    "name": obj.name,
                    "tasks": [(k, v) for k, v in obj.tasks.items()],
                }
            }
        else:
            return obj


class BatchDeserializer:
    constructor_mapping = {
        "remote_batch": RemoteBatch,
        "local_batch": LocalBatch,
        "quera_task": QuEraTask,
        "braket_task": BraketTask,
        "braket_emulator_task": BraketEmulatorTask,
        "quera_task_specification": QuEraTaskSpecification,
        "braket_task_specification": BraketTaskSpecification,
        "parallel_decoder": ParallelDecoder,
        "task_result_ir": QuEraTaskResults,
        "quera_backend": QuEraBackend,
        "mock_backend": MockBackend,
        "braket_backend": BraketBackend,
    }

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def object_hook(self, obj: Dict[str, Any]):
        if len(obj) == 1:
            ((head, body),) = obj.items()

            if head in self.constructor_mapping:
                if "tasks" in body:
                    body["tasks"] = OrderedDict(body["tasks"])
                elif head in ["quera_backend", "braket_backend", "mock_backend"]:
                    body = {**body, **self._kwargs}

                return self.constructor_mapping[head](**body)

        return obj


def save_batch(
    filename_or_io: Union[str, TextIO], batch: Union[RemoteBatch, LocalBatch]
) -> None:
    """save batch to json file or string io

    Args:
        filename_or_io (Union[str, TextIO]): filename or file object pointing to
        json file.
        batch (Union[RemoteBatch, LocalBatch]): batch object to save.
    """
    if isinstance(filename_or_io, str):
        with open(filename_or_io, "w") as f:
            json.dump(batch, f, cls=BatchSerializer)
    else:
        json.dump(batch, filename_or_io, cls=BatchSerializer)


def load_batch(
    filename_or_io: Union[str, TextIO], *backend_args, **backend_kwargs
) -> Union[RemoteBatch, LocalBatch]:
    """load batch from json file or string io to batch object

    Args:
        filename_or_io (Union[str, TextIO]): filename or file object pointing to
        json file.
        *backend_args: args to pass to backend construction.
        **backend_kwargs: kwargs to pass to backend construction.

    Returns:
        Union[RemoteBatch, LocalBatch]: the resulting batch object

    Note:
        The backend args are not always required for `LocalBatch` objects, but
        for `RemoteBatch` objects they are required.
    """
    deserializer = BatchDeserializer(*backend_args, **backend_kwargs)
    if isinstance(filename_or_io, str):
        with open(filename_or_io, "r") as f:
            return json.load(f, object_hook=deserializer.object_hook)
    else:
        return json.load(filename_or_io, object_hook=deserializer.object_hook)



from pydantic.dataclasses import dataclass
from typing import Dict, List, Tuple, Union
from decimal import Decimal


ParamType = Union[Decimal, List[Decimal]]


@dataclass(frozen=True)
class Params:
    static_params: Dict[str, ParamType]
    batch_params: List[Dict[str, ParamType]]
    args_list: Tuple[str, ...]

    def parse_args(self, *args) -> Dict[str, Decimal]:
        if len(args) != len(self.args_list):
            raise ValueError(
                f"Expected {len(self.args_list)} arguments, got {len(args)}."
            )

        args = tuple(map(Decimal, map(str, args)))
        return dict(zip(self.args_list, args))

    def batch_assignments(self, *args) -> List[Dict[str, ParamType]]:
        flattened_args = self.parse_args(*args)
        return [{**flattened_args, **batch} for batch in self.batch_params]

    def __str__(self):
        out = ""
        out += "> Static params:\n"
        for var, litrl in self.static_params.items():
            out += f" :: {var} \n      => {litrl}\n"

        out += "\n> Batch params:\n"
        for lid, pair in enumerate(self.batch_params):
            out += f"- batch {lid}:\n"
            for var, litrl in pair.items():
                out += f"   :: {var}\n       => {litrl}\n"

        out += "\n> Arguments:\n"
        out += "  " + repr(self.args_list)

        return out



from bloqade.ir.location import (
    AtomArrangement,
    Chain,
    Square,
    Rectangular,
    Honeycomb,
    Triangular,
    Lieb,
    Kagome,
    ListOfLocations,
    ParallelRegister,
)
from bloqade.ir.location.base import LocationInfo
from typing import Union, Any


AstType = Union[AtomArrangement, ParallelRegister]


class LocationVisitor:
    def visit_location_info(self, ast: LocationInfo) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for LocationInfo"
        )

    def visit_chain(self, ast: Chain) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Chain"
        )

    def visit_square(self, ast: Square) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Square"
        )

    def visit_rectangular(self, ast: Rectangular) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Rectangular"
        )

    def visit_honeycomb(self, ast: Honeycomb) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Honeycomb"
        )

    def visit_triangular(self, ast: Triangular) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Triangular"
        )

    def visit_lieb(self, ast: Lieb) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Lieb"
        )

    def visit_kagome(self, ast: Kagome) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Kagome"
        )

    def visit_list_of_locations(self, ast: ListOfLocations) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for ListOfLocations"
        )

    def visit_parallel_register(self, ast: ParallelRegister) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for ParallelRegister"
        )

    def visit(self, ast: AstType) -> Any:
        if isinstance(ast, LocationInfo):
            return self.visit_location_info(ast)
        elif isinstance(ast, Chain):
            return self.visit_chain(ast)
        elif isinstance(ast, Square):
            return self.visit_square(ast)
        elif isinstance(ast, Rectangular):
            return self.visit_rectangular(ast)
        elif isinstance(ast, Honeycomb):
            return self.visit_honeycomb(ast)
        elif isinstance(ast, Triangular):
            return self.visit_triangular(ast)
        elif isinstance(ast, Lieb):
            return self.visit_lieb(ast)
        elif isinstance(ast, Kagome):
            return self.visit_kagome(ast)
        elif isinstance(ast, ListOfLocations):
            return self.visit_list_of_locations(ast)
        elif isinstance(ast, ParallelRegister):
            return self.visit_parallel_register(ast)
        else:
            raise NotImplementedError(
                f"No visitor method implemented in {self.__class__} for {ast}"
            )



from bloqade.ir import Linear
from bloqade import start

wf = Linear(start=1.0, stop="x", duration=3.0)

seq = (
    start.rydberg.rabi.amplitude.location(1)
    .linear(start=1.0, stop=2.0, duration="x")
    .location(2)
    .linear(start=1.0, stop=2.0, duration="x")
    .constant(1.0, 3.0)
    .uniform.linear(start=1.0, stop=2.0, duration="x")
    .parse_sequence()
)

seq = (
    start.rydberg.rabi.amplitude.uniform.apply(
        Linear(start=1.0, stop=2.0, duration="x")
    )
    .location(1)
    .linear(start=1.0, stop=2.0, duration="x")
    .parse_sequence()
)

# print(seq.seq)
# print(seq.lattice)
# print(seq.assignments)



# from bloqade import start
from bloqade import start
from bloqade.ir import rydberg, detuning
from bloqade.ir.location import Square
from bloqade.ir.visitor.analog_circuit import AnalogCircuitVisitor
from bloqade.ir.control.waveform import Linear
from bloqade.ir.control.pulse import Detuning
import pytest


wv_linear = Linear(start=0, stop=4, duration=0.1)
reg = Square(3)
prog = reg.rydberg.detuning.uniform.piecewise_constant([0.1], [0.1]).parallelize(10.0)
para_reg = prog.parse_register()
seq = prog.parse_sequence()
prog2 = start.rydberg.detuning.location(1, 2).piecewise_constant([0.1], [0.1])
seq2 = prog2.parse_sequence()


def test_base_program_visitor():
    pvis = AnalogCircuitVisitor()

    # waveform
    with pytest.raises(NotImplementedError):
        pvis.visit(wv_linear)

    # registor
    with pytest.raises(NotImplementedError):
        pvis.visit(reg)

    # parallel reg
    with pytest.raises(NotImplementedError):
        pvis.visit(para_reg)

    # sequence
    with pytest.raises(NotImplementedError):
        pvis.visit(seq)

    # pulse:
    pulse = seq.pulses[rydberg]
    with pytest.raises(NotImplementedError):
        pvis.visit(pulse)

    # field:
    field = seq.pulses[rydberg].fields[detuning]
    with pytest.raises(NotImplementedError):
        pvis.visit(field)

    # SpacialModu
    scaled_loc = list(seq2.pulses[rydberg].fields[Detuning()].drives.keys())[0]
    with pytest.raises(NotImplementedError):
        pvis.visit(scaled_loc)


def test_get_natoms():
    prog = (
        Square(5)
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, 0.9, 0.1],
            [
                -10,
                -10,
                10,
                10,
            ],
        )
        .hyperfine.detuning.uniform.piecewise_linear(
            [0.1, 0.9, 0.1],
            [
                -10,
                -10,
                10,
                10,
            ],
        )
    )

    assert prog.n_atoms == 25



from bloqade.builder.base import Builder
from bloqade.builder.typing import ScalarType
from beartype.typing import Union, List, Optional, TYPE_CHECKING
import plum

if TYPE_CHECKING:
    from bloqade.builder.spatial import Uniform, Location, Scale


class Field(Builder):
    @property
    def uniform(self) -> "Uniform":
        """
        Address all atoms as part of defining the spatial modulation component
        of a drive.

        Next steps to build your program include choosing the waveform that
        will be summed with the spatial modulation to create a drive.

        The drive by itself, or the sum of subsequent drives (created by just
        chaining the construction of drives) will become the field
        (e.g. Detuning Field, Real-Valued Rabi Amplitude/Rabi Phase Field, etc.).

        - You can now do:
            - `...uniform.linear(start, stop, duration)` : to apply a linear waveform
            - `...uniform.constant(value, duration)` : to apply a constant waveform
            - `...uniform.poly([coefficients], duration)` : to apply a
                polynomial waveform
            - `...uniform.apply(wf:bloqade.ir.Waveform)`: to apply a
            pre-defined waveform
            - `...uniform.piecewise_linear([durations], [values])`:  to apply
            a piecewise linear waveform
            - `...uniform.piecewise_constant([durations], [values])`: to apply
            a piecewise constant waveform
            - `...uniform.fn(f(t,...))`: to apply a function as a waveform

        """
        from bloqade.builder.spatial import Uniform

        return Uniform(self)

    @plum.dispatch
    def _location(self, label: int, scale: Optional[ScalarType] = None):  # noqa: F811
        from bloqade.builder.spatial import Location

        if scale is None:
            scale = 1

        return Location([label], [scale], self)

    @plum.dispatch
    def _location(
        self, labels: List[int], scales: Optional[List[ScalarType]] = None
    ):  # noqa: F811
        from bloqade.builder.spatial import Location

        if scales is None:
            scales = [1] * len(labels)

        return Location(labels, scales, self)

    def location(
        self,
        labels: Union[List[int], int],
        scales: Union[List[ScalarType], ScalarType, None] = None,
    ) -> "Location":
        """Address a single atom (or multiple) atoms.

        Address a single atom (or multiple) as part of defining the spatial
        modulation component of a drive. You can specify the atoms to target
        as a list of labels and a list of scales. The scales are used to
        multiply the waveform that is applied to the atom. You can also specify
        a single label and scale to target a single atom.

        Next steps to build your program include choosing the waveform that
        will be summed with the spatial modulation to create a drive.

        The drive by itself, or the sum of subsequent drives (created by just
        chaining the construction of drives) will become the field.
        (e.g. Detuning Field, Real-Valued Rabi Amplitude/Rabi Phase Field, etc.)

        ### Usage Example:
        ```
        >>> prog = start.add_position([(0,0),(1,4),(2,8)]).rydberg.rabi
        # to target a single atom with a waveform
        >>> one_location_prog = prog.location(0)
        # to target a single atom with a scale
        >>> one_location_prog = prog.location(0, 0.5)
        # to target multiple atoms with same waveform
        >>> multi_location_prog = prog.location([0, 2])
        # to target multiple atoms with different scales
        >>> multi_location_prog = prog.location([0, 2], [0.5, "scale"])
        ```

        - You can now do:
            - `...location(labels, scales).linear(start, stop, duration)` : to apply
                a linear waveform
            - `...location(labels, scales).constant(value, duration)` : to apply
                a constant waveform
            - `...location(labels, scales).poly([coefficients], duration)` : to apply
                a polynomial waveform
            - `...location(labels, scales).apply(wf:bloqade.ir.Waveform)`: to apply
                a pre-defined waveform
            - `...location(labels, scales).piecewise_linear([durations], [values])`:
                to apply
                a piecewise linear waveform
            - `...location(labels, scales).piecewise_constant([durations], [values])`:
                to apply
                a piecewise constant waveform
            - `...location(labels, scales).fn(f(t,..))`: to apply a function as a
                waveform

        """
        return self._location(labels, scales)

    def scale(self, coeffs: Union[str, List[ScalarType]]) -> "Scale":
        """
        Address all the atoms scaling each atom with an element of the list
        or define a variable name for the scale list to be assigned later by
        defining a `name` and using `assign` or `batch_assign` later.

        Next steps to build your program include choosing the waveform that
        will be summed with the spatial modulation to create a drive.

        The drive by itself, or the sum of subsequent drives (created by just
        chaining the construction of drives) will become the field
        (e.g. Detuning Field, Real-Valued Rabi Amplitude/Rabi Phase Field, etc.)

        ### Usage Example:
        ```
        >>> prog = start.add_position([(0,0),(1,4),(2,8)]).rydberg.rabi

        # assign a literal list of values to scale each atom
        >>> one_location_prog = prog.scale([0.1, 0.2, 0.3])
        # assign a variable name to be assigned later
        >>> one_location_prog = prog.scale("a")
        # "a" can be assigned in the END of the program during variable assignment
        # using a list of values, indicating the scaling for each atom
        >>> single_assignment = ...assign(a = [0.1, 0.2, 0.3])
        # a list of lists, indicating a set of atoms should be targeted
        # for each task in a batch.
        >>> batch_assignment = ...batch_assign(a = [list_1, list_2, list_3,...])

        ```

        - You can now do:
            - `...scale(coeffs).linear(start, stop, duration)` : to apply
                a linear waveform
            - `...scale(coeffs).constant(value, duration)` : to apply
                a constant waveform
            - `...scale(coeffs).poly([coefficients], duration)` : to apply
                a polynomial waveform
            - `...scale(coeffs).apply(wf:bloqade.ir.Waveform)`: to apply
                a pre-defined waveform
            - `...scale(coeffs).piecewise_linear(durations, values)`:  to
                apply a piecewise linear waveform
            - `...scale(coeffs).piecewise_constant(durations, values)`: to
                apply a piecewise constant waveform
            - `...scale(coeffs).fn(f(t,..))`: to apply a function as a waveform

        """
        from bloqade.builder.spatial import Scale

        return Scale(coeffs, self)


class Detuning(Field):
    """
    This node represent detuning field of a
    specified level coupling (rydberg or hyperfine) type.


    Examples:

        - To specify detuning of rydberg coupling:

        >>> node = bloqade.start.rydberg.detuning
        >>> type(node)
        <class 'bloqade.builder.field.Detuning'>

        - To specify detuning of hyperfine coupling:

        >>> node = bloqade.start.hyperfine.detuning
        >>> type(node)
        <class 'bloqade.builder.field.Detuning'>

    Note:
        This node is a SpatialModulation node.
        See [`SpatialModulation`][bloqade.builder.field.SpatialModulation]
        for additional options.

    """

    def __bloqade_ir__(self):
        from bloqade.ir.control.pulse import detuning

        return detuning


# this is just an eye candy, thus
# it's not the actual Field object
# one can skip this node when doing
# compilation
class Rabi(Builder):
    """
    This node represent rabi field of a
    specified level coupling (rydberg or hyperfine) type.

    Examples:

        - To specify rabi of rydberg coupling:

        >>> node = bloqade.start.rydberg.rabi
        <class 'bloqade.builder.field.Rabi'>

        - To specify rabi of hyperfine coupling:

        >>> node = bloqade.start.hyperfine.rabi
        >>> type(node)
        <class 'bloqade.builder.field.Rabi'>


    """

    @property
    def amplitude(self) -> "RabiAmplitude":
        """
        Specify the real-valued Rabi Amplitude field.

        Next steps to build your program focus on specifying a spatial
        modulation.

        The spatial modulation, when coupled with a waveform, completes the
        specification of a "Drive". One or more drives can be summed together
        automatically to create a field such as the Rabi Amplitude here.

        - You can now
            - `...amplitude.uniform`: Address all atoms in the field
            - `...amplitude.location(...)`: Scale atoms by their indices
            - `...amplitude.scale(...)`: Scale each atom with a value from a
                list or assign a variable name to be assigned later

        """
        return RabiAmplitude(self)

    @property
    def phase(self) -> "RabiPhase":
        """
        Specify the real-valued Rabi Phase field.

        Next steps to build your program focus on specifying a spatial
        modulation.

        The spatial modulation, when coupled with a waveform, completes the
        specification of a "Drive". One or more drives can be summed together
        automatically to create a field such as the Rabi Phase here.

        - You can now
            - `...amplitude.uniform`: Address all atoms in the field
            - `...amplitude.location(...)`: Scale atoms by their indices
            - `...amplitude.scale(...)`: Scale each atom with a value from a
                list or assign a variable name to be assigned later

        """
        return RabiPhase(self)


class RabiAmplitude(Field):
    """
    This node represent amplitude of a rabi field.

    Examples:

        - To specify rabi amplitude of rydberg coupling:

        >>> node = bloqade.start.rydberg.rabi.amplitude
        >>> type(node)
        <class 'bloqade.builder.field.Amplitude'>

        - To specify rabi amplitude of hyperfine coupling:

        >>> node = bloqade.start.hyperfine.rabi.amplitude
        >>> type(node)
        <class 'bloqade.builder.field.Amplitude'>

    Note:
        This node is a SpatialModulation node.
        See [`SpatialModulation`][bloqade.builder.field.SpatialModulation]
        for additional options.

    """

    def __bloqade_ir__(self):
        from bloqade.ir.control.pulse import rabi

        return rabi.amplitude


class RabiPhase(Field):
    """
    This node represent phase of a rabi field.

    Examples:

        - To specify rabi phase of rydberg coupling:

        >>> node = bloqade.start.rydberg.rabi.phase
        >>> type(node)
        <class 'bloqade.builder.field.Phase'>

        - To specify rabi phase of hyperfine coupling:

        >>> node = bloqade.start.hyperfine.rabi.phase
        >>> type(node)
        <class 'bloqade.builder.field.Phase'>

    Note:
        This node is a SpatialModulation node.
        See [`SpatialModulation`][bloqade.builder.field.SpatialModulation]
        for additional options.
    """

    def __bloqade_ir__(self):
        from bloqade.ir.control.pulse import rabi

        return rabi.phase



import pytest

import bloqade.submission.braket
import bloqade.submission.ir.task_specification as task_spec

from bloqade.submission.ir.task_results import QuEraTaskStatusCode
from unittest.mock import patch
from bloqade.submission.base import ValidationError


def get_task_ir():
    return task_spec.QuEraTaskSpecification(
        nshots=10,
        lattice=task_spec.Lattice(sites=[(0, 0)], filling=[1]),
        effective_hamiltonian=task_spec.EffectiveHamiltonian(
            rydberg=task_spec.RydbergHamiltonian(
                rabi_frequency_amplitude=task_spec.RabiFrequencyAmplitude(
                    global_=task_spec.GlobalField(
                        times=[0, 1e-6, 2e-6, 3e-6, 4e-6],
                        values=[0, 15e6, 15e6, 0],
                    )
                ),
                rabi_frequency_phase=task_spec.RabiFrequencyPhase(
                    global_=task_spec.GlobalField(
                        times=[0, 4e-6],
                        values=[0, 0],
                    )
                ),
                detuning=task_spec.Detuning(
                    global_=task_spec.GlobalField(
                        times=[0, 1e-6, 2e-6, 3e-6, 4e-6],
                        values=[0, 15e6, 15e6, 0],
                    )
                ),
            )
        ),
    )


@patch("bloqade.submission.braket.AwsDevice")
def test_braket_submit(*args, **kwargs):
    task_ir = get_task_ir()

    backend = bloqade.submission.braket.BraketBackend()
    mock_aws_device = bloqade.submission.braket.AwsDevice(backend.device_arn)

    backend.submit_task(task_ir)

    mock_aws_device.run.assert_called_once()


@patch("bloqade.submission.braket.AwsDevice")
def test_add_braket_user_agent_invoked(*args, **kwargs):
    backend = bloqade.submission.braket.BraketBackend()
    expected_user_agent = f"Bloqade/{bloqade.__version__}"

    backend.device.aws_session.add_braket_user_agent.assert_called_with(
        expected_user_agent
    )


@pytest.mark.skip(
    reason="removed implementation for validation because of issue with empty queue."
)
@patch("bloqade.submission.braket.AwsDevice")
@patch("bloqade.submission.braket.AwsQuantumTask")
def test_braket_validate_task(*args, **kwargs):
    task_ir = get_task_ir()

    backend = bloqade.submission.braket.BraketBackend()
    mock_aws_device = bloqade.submission.braket.AwsDevice(backend.device_arn)
    mock_aws_device.run.return_value = bloqade.submission.braket.AwsQuantumTask(
        "task_id"
    )
    mock_aws_device.run.return_value.id = "task_id"

    # test passing validation
    backend.validate_task(task_ir)

    mock_aws_device.run.assert_called_once()
    mock_aws_device.run.return_value.cancel.assert_called_once()

    mock_aws_device.reset_mock()

    # test failing validation
    mock_aws_device = bloqade.submission.braket.AwsDevice(backend.device_arn)
    mock_aws_device.run.side_effect = Exception("ValidationException: validation error")
    with pytest.raises(ValidationError):
        backend.validate_task(task_ir)

    mock_aws_device.run.assert_called_once()

    mock_aws_device.reset_mock()

    # test failing validation
    mock_aws_device = bloqade.submission.braket.AwsDevice(backend.device_arn)
    mock_aws_device.run.side_effect = Exception("other error")
    with pytest.raises(Exception):
        backend.validate_task(task_ir)

    mock_aws_device.run.assert_called_once()


@patch("bloqade.submission.braket.AwsQuantumTask")
def test_braket_fetch(*args, **kwargs):
    backend = bloqade.submission.braket.BraketBackend()
    mock_aws_quantum_task = bloqade.submission.braket.AwsQuantumTask("task_id")

    backend.task_results("task_id")

    mock_aws_quantum_task.result.assert_called_once()


@patch("bloqade.submission.braket.AwsQuantumTask")
def test_braket_cancel(*args, **kwargs):
    backend = bloqade.submission.braket.BraketBackend()
    mock_aws_quantum_task = bloqade.submission.braket.AwsQuantumTask("task_id")

    backend.cancel_task("task_id")

    mock_aws_quantum_task.cancel.assert_called_once()


@patch("bloqade.submission.braket.AwsQuantumTask")
def test_braket_status(*args, **kwargs):
    backend = bloqade.submission.braket.BraketBackend()
    mock_aws_quantum_task = bloqade.submission.braket.AwsQuantumTask("task_id")
    mock_aws_quantum_task.state.side_effect = [
        "CREATED",
        "RUNNING",
        "COMPLETED",
        "FAILED",
        "CANCELLED",
        "QUEUED",
        "ASDLFKASLDF",
    ]

    assert backend.task_status("task_id") == QuEraTaskStatusCode.Created
    assert backend.task_status("task_id") == QuEraTaskStatusCode.Running
    assert backend.task_status("task_id") == QuEraTaskStatusCode.Completed
    assert backend.task_status("task_id") == QuEraTaskStatusCode.Failed
    assert backend.task_status("task_id") == QuEraTaskStatusCode.Cancelled
    assert backend.task_status("task_id") == QuEraTaskStatusCode.Enqueued
    with pytest.raises(ValueError):
        backend.task_status("task_id")



from bloqade.submission.base import SubmissionBackend

from bloqade.submission.ir.task_specification import (
    QuEraTaskSpecification,
)
from bloqade.submission.ir.task_results import (
    QuEraTaskResults,
    QuEraTaskStatusCode,
    QuEraShotResult,
    QuEraShotStatusCode,
)
import uuid
import numpy as np


def simulate_task_results(task: QuEraTaskSpecification, p_full=0.99, p_empty=0.01):
    natoms = len(task.lattice.sites)
    filling = task.lattice.filling

    pre_sequence_probs = np.array(
        [p_full if fill == 1 else p_empty for fill in filling]
    )
    post_sequence_probs = np.array(natoms * [0.5])

    rng = np.random.default_rng()

    shot_outputs = []
    for shot in range(task.nshots):
        pre_sequence = rng.binomial(np.ones(natoms, dtype=int), pre_sequence_probs)
        post_sequence = rng.binomial(
            np.ones(natoms, dtype=int), pre_sequence * post_sequence_probs
        )

        shot_outputs.append(
            QuEraShotResult(
                shot_status=QuEraShotStatusCode.Completed,
                pre_sequence=list(pre_sequence),
                post_sequence=list(post_sequence),
            )
        )

    return QuEraTaskResults(
        task_status=QuEraTaskStatusCode.Completed, shot_outputs=shot_outputs
    )


class MockBackend(SubmissionBackend):
    state_file: str = ".mock_state.txt"
    submission_error: bool = False

    def submit_task(self, task: QuEraTaskSpecification) -> str:
        if self.submission_error:
            raise ValueError("mock submission error")

        task_id = str(uuid.uuid4())
        task_results = simulate_task_results(task)
        with open(self.state_file, "a") as IO:
            IO.write(f"('{task_id}',{task_results.json()})\n")

        return task_id

    def task_results(self, task_id: str) -> QuEraTaskResults:
        # lazily search database for task_id
        for line in open(self.state_file, "r"):
            potential_task_id, task_results = eval(line)

            if potential_task_id == task_id:
                return QuEraTaskResults(**task_results)

        raise ValueError(f"unable to fetch results for task_id: {task_id}")

    def cancel_task(self, task_id: str):
        pass

    def task_status(self, task_id: str) -> QuEraTaskStatusCode:
        return QuEraTaskStatusCode.Completed



from bloqade.ir import (
    rydberg,
    detuning,
    Sequence,
    Field,
    Pulse,
    Uniform,
    Linear,
    Constant,
    Interval,
    NamedPulse,
)
from bloqade import cast
import bloqade.ir.analysis.assignment_scan as asn
import bloqade.codegen.hardware.quera as quer
from bloqade.ir.control.sequence import NamedSequence
from bloqade.ir.control.pulse import Slice


def cf(inlist):
    return list(map(lambda x: float(x), inlist))


def test_assignment_scan_app():
    f = Field({Uniform: Linear(start=1.0, stop=2.0, duration=3.0)})
    ps = Pulse({detuning: f})
    seq_full = Sequence({rydberg: ps})

    app = seq_full.append(seq_full)

    asgn = {"test": 40}
    scanner = asn.AssignmentScan({"test": 40})

    assert scanner.emit(app) == asgn


def test_assignment_scan_slice():
    f = Field({Uniform: Linear(start=1.0, stop=2.0, duration=3.0)})
    ps = Pulse({detuning: f})
    seq_full = Sequence({rydberg: ps})

    seq = seq_full[:1.0]

    asgn = {"test": 40}
    scanner = asn.AssignmentScan({"test": 40})

    assert scanner.emit(seq) == asgn


def test_assignment_scan_namedseq():
    f = Field({Uniform: Linear(start=1.0, stop=2.0, duration=3.0)})
    ps = Pulse({detuning: f})
    seq_full = Sequence({rydberg: ps})

    seq = NamedSequence(seq_full, "qq")

    asgn = {"test": 40}
    scanner = asn.AssignmentScan({"test": 40})

    assert scanner.emit(seq) == asgn


def test_assignment_scan_waveform():
    wv = Linear(start=1.0, stop=2.0, duration=3.0)
    wf = wv + wv

    f = Field({Uniform: wf})
    ps = ps = NamedPulse("qq", Pulse({detuning: f}))

    seq_full = Sequence({rydberg: ps})

    asgn = {"test": 40}
    scanner = asn.AssignmentScan(asgn)

    assert scanner.emit(seq_full) == asgn


def test_assignment_scan_pulse_app():
    wv = Linear(start=1.0, stop=2.0, duration=3.0)

    f = Field({Uniform: wv})
    ps = Pulse({detuning: f})
    pss = ps.append(ps)
    seq_full = Sequence({rydberg: pss})

    asgn = {"test": 40}
    scanner = asn.AssignmentScan(asgn)

    assert scanner.emit(seq_full) == asgn


def test_assignment_scan_pulse_slice():
    wv = Linear(start=1.0, stop=2.0, duration=3.0)

    f = Field({Uniform: wv})
    ps = Pulse({detuning: f})
    pss = Slice(pulse=ps, interval=Interval(None, cast(1.0)))
    seq_full = Sequence({rydberg: pss})

    asgn = {"test": 40}
    scanner = asn.AssignmentScan(asgn)

    assert scanner.emit(seq_full) == asgn


def test_plin_codegen_slice():
    wv = Linear(start=0, stop=1.0, duration=1.0)
    wv = wv.append(Linear(start=1.0, stop=2.0, duration=2.0))

    asgn = {"test": 40}
    scanner = quer.PiecewiseLinearCodeGen(asgn)

    wf = wv[:0.5]
    pwl = scanner.emit(wf)
    assert (cf(pwl.times), cf(pwl.values)) == ([0, 0.5], [0, 0.5])

    wf2 = wv[0:1]
    pwl = scanner.emit(wf2)
    assert (cf(pwl.times), cf(pwl.values)) == ([0, 1], [0, 1])

    wf3 = wv[0.7:1]
    pwl = scanner.emit(wf3)
    assert (cf(pwl.times), cf(pwl.values)) == ([0, 0.3], [0.7, 1])

    wf4 = wv[0:0.6]
    pwl = scanner.emit(wf4)
    assert (cf(pwl.times), cf(pwl.values)) == ([0, 0.6], [0.0, 0.6])

    wf5 = wv[0.2:0.6]
    pwl = scanner.emit(wf5)
    assert (cf(pwl.times), cf(pwl.values)) == ([0, 0.4], [0.2, 0.6])

    wf6 = wv[0.6:1.5]
    pwl = scanner.emit(wf6)
    assert (cf(pwl.times), cf(pwl.values)) == ([0, 0.4, 0.9], [0.6, 1.0, 1.25])

    wf7 = wv[1.0:1.5]
    pwl = scanner.emit(wf7)
    assert (cf(pwl.times), cf(pwl.values)) == ([0, 0.5], [1.0, 1.25])

    wf8 = wv[1.0:3.0]
    pwl = scanner.emit(wf8)
    assert (cf(pwl.times), cf(pwl.values)) == ([0, 2.0], [1.0, 2.0])

    wf9 = wv[1.5:3.0]
    pwl = scanner.emit(wf9)
    assert (cf(pwl.times), cf(pwl.values)) == ([0, 1.5], [1.25, 2.0])


def test_pconst_codegen_slice():
    wv = Constant(value=1.0, duration=1.0)
    wv = wv.append(Constant(value=2.0, duration=1.5))

    asgn = {"test": 40}
    scanner = quer.PiecewiseConstantCodeGen(asgn)

    wf = wv[:1.3]
    pwc = scanner.emit(wf)
    assert (cf(pwc.times), cf(pwc.values)) == ([0, 1.0, 1.3], [1.0, 2.0, 2.0])

    wf2 = wv[0:1]
    pwc = scanner.emit(wf2)
    assert (cf(pwc.times), cf(pwc.values)) == ([0, 1.0], [1.0, 1.0])

    wf3 = wv[0.7:1.2]
    pwc = scanner.emit(wf3)
    assert (cf(pwc.times), cf(pwc.values)) == ([0, 0.3, 0.5], [1.0, 2.0, 2.0])

    wf4 = wv[0.7:1]
    pwc = scanner.emit(wf4)
    assert (cf(pwc.times), cf(pwc.values)) == ([0, 0.3], [1.0, 1])

    wf4 = wv[0:0.6]
    pwc = scanner.emit(wf4)
    assert (cf(pwc.times), cf(pwc.values)) == ([0, 0.6], [1.0, 1.0])

    wf5 = wv[0.2:0.6]
    pwc = scanner.emit(wf5)
    assert (cf(pwc.times), cf(pwc.values)) == ([0, 0.4], [1.0, 1.0])

    wf6 = wv[0:0]
    pwc = scanner.emit(wf6)
    assert (cf(pwc.times), cf(pwc.values)) == ([0, 0], [0, 0])

    wf8 = wv[1:1]
    pwc = scanner.emit(wf8)
    assert (cf(pwc.times), cf(pwc.values)) == ([0, 0], [0, 0])

    wf7 = wv[1.3:1.3]
    pwc = scanner.emit(wf7)
    assert (cf(pwc.times), cf(pwc.values)) == ([0, 0], [0, 0])

    wf9 = wv[1:2.5]
    pwc = scanner.emit(wf9)
    assert (cf(pwc.times), cf(pwc.values)) == ([0, 1.5], [2.0, 2.0])

    wf10 = wv[1.5:2.5]
    pwc = scanner.emit(wf10)
    assert (cf(pwc.times), cf(pwc.values)) == ([0, 1.0], [2.0, 2.0])

    wf11 = wv[1:1.5]
    pwc = scanner.emit(wf11)
    assert (cf(pwc.times), cf(pwc.values)) == ([0, 0.5], [2.0, 2.0])



from pydantic import PrivateAttr
from bloqade.submission.base import SubmissionBackend, ValidationError
from bloqade.submission.quera_api_client.api import QueueApi
from bloqade.submission.ir.task_specification import (
    QuEraTaskSpecification,
)
from bloqade.submission.ir.capabilities import QuEraCapabilities
from bloqade.submission.ir.task_results import (
    QuEraTaskResults,
    QuEraTaskStatusCode,
)
from typing import Optional


class QuEraBackend(SubmissionBackend):
    api_hostname: str
    qpu_id: str
    api_stage: str = "v0"
    virtual_queue: Optional[str] = None
    proxy: Optional[str] = None
    # Sigv4Request arguments
    region: Optional[str] = None
    access_key: Optional[str] = None
    secret_key: Optional[str] = None
    session_token: Optional[str] = None
    session_expires: Optional[int] = None
    role_arn: Optional[str] = None
    role_session_name: Optional[str] = None
    profile: Optional[str] = None
    _queue_api: Optional[QueueApi] = PrivateAttr(None)

    @property
    def queue_api(self):
        if self._queue_api is None:
            kwargs = {k: v for k, v in self.__dict__.items() if v is not None}
            self._queue_api = QueueApi(**kwargs)

        return self._queue_api

    def get_capabilities(self) -> QuEraCapabilities:
        try:
            return QuEraCapabilities(**self.queue_api.get_capabilities())
        except BaseException:
            return super().get_capabilities()

    def submit_task(self, task_ir: QuEraTaskSpecification) -> str:
        return self.queue_api.post_task(
            task_ir.json(by_alias=True, exclude_none=True, exclude_unset=True)
        )

    def task_results(self, task_id: str) -> QuEraTaskResults:
        return QuEraTaskResults(**self.queue_api.poll_task_results(task_id))

    def cancel_task(self, task_id: str):
        self.queue_api.cancel_task_in_queue(task_id)

    def task_status(self, task_id: str) -> QuEraTaskStatusCode:
        return_body = self.queue_api.get_task_status_in_queue(task_id)
        return QuEraTaskStatusCode(return_body)

    def validate_task(self, task_ir: QuEraTaskSpecification):
        try:
            self.queue_api.validate_task(
                task_ir.json(by_alias=True, exclude_none=True, exclude_unset=True)
            )
        except QueueApi.ValidationError as e:
            raise ValidationError(str(e))

    def update_credential(
        self, access_key: str = None, secret_key: str = None, session_token: str = None
    ):
        if secret_key is not None:
            self.secret_key = secret_key
        if access_key is not None:
            self.access_key = access_key
        if session_token is not None:
            self.session_token = session_token



from typing import Any
import bloqade.ir.analog_circuit as analog_circuit
import bloqade.ir.control.sequence as sequence
from bloqade.ir.visitor.analog_circuit import AnalogCircuitVisitor


class IsHyperfineSequence(AnalogCircuitVisitor):
    def __init__(self):
        self.is_hyperfine = False

    def visit_analog_circuit(self, ast: analog_circuit.AnalogCircuit) -> Any:
        self.visit(ast.sequence)

    def visit_append_sequence(self, ast: sequence.Append) -> Any:
        list(map(self.visit, ast.sequences))

    def visit_slice_sequence(self, ast: sequence.Slice) -> Any:
        self.visit(ast.sequence)

    def visit_named_sequence(self, ast: sequence.NamedSequence) -> Any:
        self.visit(ast.sequence)

    def visit_sequence(self, ast: sequence.Sequence) -> Any:
        self.is_hyperfine = self.is_hyperfine or sequence.hyperfine in ast.pulses

    def emit(self, ast: analog_circuit.AnalogCircuit) -> bool:
        self.visit(ast)
        return self.is_hyperfine



from bloqade.ir.location import (
    AtomArrangement,
    Chain,
    Square,
    Rectangular,
    Honeycomb,
    Triangular,
    Lieb,
    Kagome,
    ListOfLocations,
)


__all__ = [
    "AtomArrangement",
    "Chain",
    "Square",
    "Rectangular",
    "Honeycomb",
    "Triangular",
    "Lieb",
    "Kagome",
    "ListOfLocations",
]



from bloqade.ir import Linear
import bloqade.ir.location as location

(
    location.Square(3)
    .rydberg.detuning.uniform.apply(Linear(start=1.0, stop="x", duration=3.0))
    .location(2)
    .scale(3.0)
    .apply(Linear(start=1.0, stop="x", duration=3.0))
)

(
    location.Square(3)
    .rydberg.detuning.uniform.apply(Linear(start=1.0, stop="x", duration=3.0))
    .location(2)
    .scale(3.0)
    .apply(Linear(start=1.0, stop="x", duration=3.0))
)



from bokeh.models import (
    ColumnDataSource,
    Range1d,
    HoverTool,
)
from bokeh.plotting import figure
import numpy as np


def get_atom_arrangement_figure(atom_arng_ir, fig_kwargs=None, **assignments):
    import bloqade.ir.location as ir_loc

    """obtain a figure object from the atom arrangement."""
    xs_filled, ys_filled, labels_filled = [], [], []
    xs_vacant, ys_vacant, labels_vacant = [], [], []
    x_min = np.inf
    x_max = -np.inf
    y_min = np.inf
    y_max = -np.inf
    for idx, location_info in enumerate(atom_arng_ir.enumerate()):
        (x_var, y_var) = location_info.position
        (x, y) = (float(x_var(**assignments)), float(y_var(**assignments)))
        x_min = min(x, x_min)
        y_min = min(y, y_min)
        x_max = max(x, x_max)
        y_max = max(y, y_max)
        if location_info.filling is ir_loc.base.SiteFilling.filled:
            xs_filled.append(x)
            ys_filled.append(y)
            labels_filled.append(idx)
        else:
            xs_vacant.append(x)
            ys_vacant.append(y)
            labels_vacant.append(idx)

    if atom_arng_ir.n_atoms > 0:
        length_scale = max(y_max - y_min, x_max - x_min, 1)
    else:
        length_scale = 1

    source_filled = ColumnDataSource(
        data=dict(_x=xs_filled, _y=ys_filled, _labels=labels_filled)
    )
    source_vacant = ColumnDataSource(
        data=dict(_x=xs_vacant, _y=ys_vacant, _labels=labels_vacant)
    )
    source_all = ColumnDataSource(
        data=dict(
            _x=xs_vacant + xs_filled,
            _y=ys_vacant + ys_filled,
            _labels=labels_vacant + labels_filled,
        )
    )
    hover = HoverTool()
    hover.tooltips = [
        ("(x,y)", "(@_x, @_y)"),
        ("index: ", "@_labels"),
    ]

    ## remove box_zoom since we don't want to change the scale
    if fig_kwargs is None:
        fig_kwargs = {}

    p = figure(
        **fig_kwargs,
        width=400,
        height=400,
        tools="wheel_zoom,reset, undo, redo, pan",
        toolbar_location="above",
    )
    p.x_range = Range1d(x_min - 1, x_min + length_scale + 1)
    p.y_range = Range1d(y_min - 1, y_min + length_scale + 1)

    p.circle(
        "_x", "_y", source=source_filled, radius=0.015 * length_scale, fill_alpha=1
    )
    p.circle(
        "_x",
        "_y",
        source=source_vacant,
        radius=0.015 * length_scale,
        fill_alpha=0.25,
        color="grey",
        line_width=0.2 * length_scale,
    )

    p.circle(
        "_x",
        "_y",
        source=source_all,
        radius=0,  # in the same unit as the data
        fill_alpha=0.5,
        line_width=0.15 * length_scale,
        visible=True,  # display by default
        name="Brad",
    )
    p.add_tools(hover)

    return p



from bloqade.emulate.ir.atom_type import TwoLevelAtom
from bloqade.emulate.ir.emulator import (
    EmulatorProgram,
    Register,
    RabiOperatorData,
    RabiTerm,
    Fields,
    DetuningTerm,
    DetuningOperatorData,
    RabiOperatorType,
)
from bloqade.codegen.emulator_ir import (
    EmulatorProgramCodeGen,
    CompiledWaveform,
    LevelCoupling,
)
from bloqade import start
from bloqade.ir.control.sequence import rydberg
from bloqade.ir.control.pulse import detuning, rabi
from bloqade.ir.control.field import Location, RunTimeVector, ScaledLocations, Uniform
from decimal import Decimal

from bloqade.ir.scalar import cast


def test_codegen_global_detuning():
    program = (
        start.add_position((0, 0))
        .add_position((0, 5))
        .rydberg.detuning.uniform.piecewise_linear([0.1, 0.8, 0.1], [-10, -10, 10, 10])
        .parse_circuit()
    )

    detuning_waveform = (
        program.sequence.pulses[rydberg].fields[detuning].drives[Uniform]
    )

    compiled_waveform = CompiledWaveform({}, detuning_waveform)

    geometry = Register(
        TwoLevelAtom,
        [(Decimal("0"), Decimal("0")), (Decimal("0"), Decimal("5"))],
        Decimal("0"),
    )

    detuning_term = DetuningTerm(
        DetuningOperatorData({i: Decimal("1") for i in range(len(geometry))}),
        compiled_waveform,
    )

    rydberg_drive = Fields(detuning=[detuning_term], rabi=[])

    drives = {LevelCoupling.Rydberg: rydberg_drive}

    expected_emulator_ir = EmulatorProgram(
        register=geometry, duration=1.0, pulses=drives
    )

    emulator_ir = EmulatorProgramCodeGen().emit(program)

    print(emulator_ir)
    print(expected_emulator_ir)

    assert emulator_ir == expected_emulator_ir


def test_codegen_global_detuning_and_rabi():
    program = (
        start.add_position((0, 0))
        .add_position((0, 5))
        .rydberg.detuning.uniform.piecewise_linear([0.1, 0.8, 0.1], [-10, -10, 10, 10])
        .amplitude.uniform.piecewise_linear([0.1, 0.8, 0.1], [0, 10, 10, 0])
        .parse_circuit()
    )

    detuning_waveform = (
        program.sequence.pulses[rydberg].fields[detuning].drives[Uniform]
    )
    amplitude_waveform = (
        program.sequence.pulses[rydberg].fields[rabi.amplitude].drives[Uniform]
    )

    compiled_detuning = CompiledWaveform({}, detuning_waveform)
    compiled_amplitude = CompiledWaveform({}, amplitude_waveform)

    geometry = Register(
        TwoLevelAtom,
        [(Decimal("0"), Decimal("0")), (Decimal("0"), Decimal("5"))],
        Decimal("0"),
    )

    detuning_term = DetuningTerm(
        DetuningOperatorData({i: Decimal("1") for i in range(len(geometry))}),
        compiled_detuning,
    )

    rabi_term = RabiTerm(
        RabiOperatorData(
            RabiOperatorType.RabiSymmetric,
            {i: Decimal("1") for i in range(len(geometry))},
        ),
        compiled_amplitude,
        None,
    )

    rydberg_drive = Fields(detuning=[detuning_term], rabi=[rabi_term])

    drives = {LevelCoupling.Rydberg: rydberg_drive}

    expected_emulator_ir = EmulatorProgram(
        register=geometry, duration=1.0, pulses=drives
    )

    emulator_ir = EmulatorProgramCodeGen().emit(program)

    print(emulator_ir)
    print(expected_emulator_ir)

    assert emulator_ir == expected_emulator_ir


def test_codegen_detuning_max_terms():
    program = (
        start.add_position((0, 0))
        .add_position((0, 5))
        .rydberg.detuning.uniform.piecewise_linear([0.1, 0.8, 0.1], [-10, -10, 10, 10])
        .scale("mask_1")
        .piecewise_linear([0.1, 0.8, 0.1], [-20, -20, 10, 10])
        .location(0, 0.4)
        .piecewise_linear([0.1, 0.8, 0.1], [-10, -10, 20, 20])
        .parse_circuit()
    )
    mask_1_value = [0.5, 2.0]

    mask_1 = RunTimeVector("mask_1")
    location_0 = ScaledLocations({Location(0): cast(0.4)})
    uniform_wf = program.sequence.pulses[rydberg].fields[detuning].drives[Uniform]
    mask_1_wf = program.sequence.pulses[rydberg].fields[detuning].drives[mask_1]
    location_0_wf = program.sequence.pulses[rydberg].fields[detuning].drives[location_0]

    wf_0 = (1.0 * uniform_wf + mask_1_value[0] * mask_1_wf) + 0.4 * location_0_wf
    wf_1 = 1.0 * uniform_wf + mask_1_value[1] * mask_1_wf

    assignments = {"mask_1": mask_1_value}

    compiled_wf_0 = CompiledWaveform(assignments, wf_0)
    compiled_wf_1 = CompiledWaveform(assignments, wf_1)

    geometry = Register(
        TwoLevelAtom,
        [(Decimal("0"), Decimal("0")), (Decimal("0"), Decimal("5"))],
        Decimal("0"),
    )

    detuning_term_0 = DetuningTerm(
        DetuningOperatorData({0: Decimal("1")}),
        compiled_wf_0,
    )

    detuning_term_1 = DetuningTerm(
        DetuningOperatorData({1: Decimal("1")}),
        compiled_wf_1,
    )

    rydberg_drive = Fields(detuning=[detuning_term_0, detuning_term_1], rabi=[])
    drives = {LevelCoupling.Rydberg: rydberg_drive}

    expected_emulator_ir = EmulatorProgram(
        register=geometry, duration=1.0, pulses=drives
    )

    emulator_ir = EmulatorProgramCodeGen(assignments=assignments).emit(program)

    print(emulator_ir)
    print(expected_emulator_ir)

    assert emulator_ir == expected_emulator_ir


def test_codegen_rabi_max_terms():
    program = (
        start.add_position((0, 0))
        .add_position((0, 5))
        .rydberg.rabi.amplitude.uniform.piecewise_linear(
            [0.1, 0.8, 0.1], [-10, -10, 10, 10]
        )
        .scale("mask_1")
        .piecewise_linear([0.1, 0.8, 0.1], [-20, -20, 10, 10])
        .location(0, 0.4)
        .piecewise_linear([0.1, 0.8, 0.1], [-10, -10, 20, 20])
        .parse_circuit()
    )
    mask_1_value = [0.5, 2.0]

    mask_1 = RunTimeVector("mask_1")
    location_0 = ScaledLocations({Location(0): cast(0.4)})
    uniform_wf = program.sequence.pulses[rydberg].fields[rabi.amplitude].drives[Uniform]
    mask_1_wf = program.sequence.pulses[rydberg].fields[rabi.amplitude].drives[mask_1]
    location_0_wf = (
        program.sequence.pulses[rydberg].fields[rabi.amplitude].drives[location_0]
    )

    wf_0 = (1.0 * uniform_wf + mask_1_value[0] * mask_1_wf) + 0.4 * location_0_wf
    wf_1 = 1.0 * uniform_wf + mask_1_value[1] * mask_1_wf

    assignments = {"mask_1": mask_1_value}

    compiled_wf_0 = CompiledWaveform(assignments, wf_0)
    compiled_wf_1 = CompiledWaveform(assignments, wf_1)

    geometry = Register(
        TwoLevelAtom,
        [(Decimal("0"), Decimal("0")), (Decimal("0"), Decimal("5"))],
        Decimal("0"),
    )

    rabi_term_0 = RabiTerm(
        RabiOperatorData(
            RabiOperatorType.RabiSymmetric,
            {0: Decimal("1")},
        ),
        compiled_wf_0,
        None,
    )

    rabi_term_1 = RabiTerm(
        RabiOperatorData(
            RabiOperatorType.RabiSymmetric,
            {1: Decimal("1")},
        ),
        compiled_wf_1,
        None,
    )

    rydberg_drive = Fields(detuning=[], rabi=[rabi_term_0, rabi_term_1])

    drives = {LevelCoupling.Rydberg: rydberg_drive}

    expected_emulator_ir = EmulatorProgram(
        register=geometry, duration=1.0, pulses=drives
    )

    emulator_ir = EmulatorProgramCodeGen(assignments=assignments).emit(program)

    print(emulator_ir)
    print(expected_emulator_ir)

    assert emulator_ir == expected_emulator_ir


def test_codegen_rabi_uniform_phase():
    program = (
        start.add_position((0, 0))
        .add_position((0, 5))
        .rydberg.rabi.amplitude.uniform.piecewise_linear(
            [0.1, 0.8, 0.1], [-10, -10, 10, 10]
        )
        .rydberg.rabi.phase.uniform.piecewise_linear(
            [0.1, 0.8, 0.1], [-10, -10, 10, 10]
        )
        .parse_circuit()
    )

    uniform_amp = (
        program.sequence.pulses[rydberg].fields[rabi.amplitude].drives[Uniform]
    )
    uniform_phase = program.sequence.pulses[rydberg].fields[rabi.phase].drives[Uniform]

    assignments = {}

    compiled_amp = CompiledWaveform(assignments, uniform_amp)
    compiled_phase = CompiledWaveform(assignments, uniform_phase)

    geometry = Register(
        TwoLevelAtom,
        [(Decimal("0"), Decimal("0")), (Decimal("0"), Decimal("5"))],
        Decimal("0"),
    )

    rabi_term = RabiTerm(
        RabiOperatorData(
            RabiOperatorType.RabiAsymmetric,
            {i: Decimal("1") for i in range(len(geometry))},
        ),
        compiled_amp,
        compiled_phase,
    )

    rydberg_drive = Fields(detuning=[], rabi=[rabi_term])

    drives = {LevelCoupling.Rydberg: rydberg_drive}

    expected_emulator_ir = EmulatorProgram(
        register=geometry, duration=1.0, pulses=drives
    )

    emulator_ir = EmulatorProgramCodeGen(assignments=assignments).emit(program)

    print(emulator_ir)
    print(expected_emulator_ir)

    assert emulator_ir == expected_emulator_ir


def test_codegen_uniform_phase_rabi_max_terms():
    program = (
        start.add_position((0, 0))
        .add_position((0, 5))
        .rydberg.rabi.amplitude.uniform.piecewise_linear(
            [0.1, 0.8, 0.1], [-10, -10, 10, 10]
        )
        .scale("mask_1")
        .piecewise_linear([0.1, 0.8, 0.1], [-20, -20, 10, 10])
        .location(0, 0.4)
        .piecewise_linear([0.1, 0.8, 0.1], [-10, -10, 20, 20])
        .phase.uniform.piecewise_linear([0.1, 0.8, 0.1], [-10, -10, 10, 10])
        .parse_circuit()
    )

    mask_1_value = [0.5, 2.0]
    mask_1 = RunTimeVector("mask_1")
    location_0 = ScaledLocations({Location(0): cast(0.4)})

    uniform_amp = (
        program.sequence.pulses[rydberg].fields[rabi.amplitude].drives[Uniform]
    )
    mask_1_amp = program.sequence.pulses[rydberg].fields[rabi.amplitude].drives[mask_1]
    location_0_amp = (
        program.sequence.pulses[rydberg].fields[rabi.amplitude].drives[location_0]
    )
    phase_wf = program.sequence.pulses[rydberg].fields[rabi.phase].drives[Uniform]

    wf_0 = (1.0 * uniform_amp + mask_1_value[0] * mask_1_amp) + 0.4 * location_0_amp
    wf_1 = 1.0 * uniform_amp + mask_1_value[1] * mask_1_amp
    wf_0_phase = 1.0 * phase_wf
    wf_1_phase = 1.0 * phase_wf

    assignments = {"mask_1": mask_1_value}

    compiled_wf_0 = CompiledWaveform(assignments, wf_0)
    compiled_wf_1 = CompiledWaveform(assignments, wf_1)
    compiled_wf_0_phase = CompiledWaveform(assignments, wf_0_phase)
    compiled_wf_1_phase = CompiledWaveform(assignments, wf_1_phase)

    geometry = Register(
        TwoLevelAtom,
        [(Decimal("0"), Decimal("0")), (Decimal("0"), Decimal("5"))],
        Decimal("0"),
    )

    rabi_term_0 = RabiTerm(
        RabiOperatorData(
            RabiOperatorType.RabiAsymmetric,
            {0: Decimal("1")},
        ),
        compiled_wf_0,
        compiled_wf_0_phase,
    )

    rabi_term_1 = RabiTerm(
        RabiOperatorData(
            RabiOperatorType.RabiAsymmetric,
            {1: Decimal("1")},
        ),
        compiled_wf_1,
        compiled_wf_1_phase,
    )

    rydberg_drive = Fields(detuning=[], rabi=[rabi_term_0, rabi_term_1])

    drives = {LevelCoupling.Rydberg: rydberg_drive}

    expected_emulator_ir = EmulatorProgram(
        register=geometry, duration=1.0, pulses=drives
    )

    emulator_ir = EmulatorProgramCodeGen(assignments=assignments).emit(program)

    print(emulator_ir)
    print(expected_emulator_ir)

    assert emulator_ir == expected_emulator_ir


if __name__ == "__main__":
    test_codegen_uniform_phase_rabi_max_terms()



# from bloqade import start
from bloqade.ir import rabi
from bloqade.ir import cast
from bloqade.ir import Interpolation
from bloqade.ir import Interval
from bloqade.ir.visitor.waveform import WaveformVisitor
from bloqade.ir.control.waveform import (
    GaussianKernel,
    Alignment,
    Linear,
    Constant,
    Append,
    Slice,
    Negative,
    Scale,
    Add,
    Record,
    PythonFn,
    Sample,
    Poly,
    Smooth,
    AlignedWaveform,
)
import pytest


wv_linear = Linear(start=0, stop=4, duration=0.1)
wv_constant = Constant(value=0, duration=0.1)
wv_append = Append([wv_linear, wv_constant])
wv_slice = Slice(waveform=wv_linear, interval=Interval(cast(0.1), cast(0.2)))
wv_negative = Negative(wv_linear)
wv_scale = Scale(cast(4), wv_linear)
wv_add = Add(wv_linear, wv_constant)
wv_record = Record(wv_linear, cast("tst"))
wv_python = PythonFn.create(lambda x: x**2, duration=0.5)
wv_sample = Sample(wv_linear, Interpolation("linear"), dt=cast(0.5))
wv_poly = Poly(coeffs=[0.1, 0.2], duration=0.5)
wv_smooth = Smooth(radius=cast(0.5), kernel=GaussianKernel, waveform=wv_linear)
wv_align = AlignedWaveform(wv_linear, Alignment("left_aligned"), cast(0.5))


def test_base_wvfm_visitor():
    wvis = WaveformVisitor()

    with pytest.raises(NotImplementedError):
        wvis.visit(wv_align)

    with pytest.raises(NotImplementedError):
        wvis.visit(wv_linear)

    with pytest.raises(NotImplementedError):
        wvis.visit(wv_constant)

    with pytest.raises(NotImplementedError):
        wvis.visit(wv_append)

    with pytest.raises(NotImplementedError):
        wvis.visit(wv_slice)

    with pytest.raises(NotImplementedError):
        wvis.visit(wv_negative)

    with pytest.raises(NotImplementedError):
        wvis.visit(wv_scale)

    with pytest.raises(NotImplementedError):
        wvis.visit(wv_add)

    with pytest.raises(NotImplementedError):
        wvis.visit(wv_record)

    with pytest.raises(NotImplementedError):
        wvis.visit(wv_python)

    with pytest.raises(NotImplementedError):
        wvis.visit(wv_sample)

    with pytest.raises(NotImplementedError):
        wvis.visit(wv_poly)

    with pytest.raises(NotImplementedError):
        wvis.visit(wv_smooth)

    ## not waveform ast
    with pytest.raises(ValueError):
        wvis.visit(rabi)



from bloqade.serialize import Serializer
from bloqade.submission.mock import MockBackend
from bloqade.task.base import Geometry
from bloqade.task.base import RemoteTask

from bloqade.submission.base import ValidationError
from bloqade.submission.ir.task_results import QuEraTaskResults, QuEraTaskStatusCode
from bloqade.submission.ir.task_specification import QuEraTaskSpecification
from bloqade.submission.ir.parallel import ParallelDecoder
from bloqade.submission.quera import QuEraBackend

from beartype.typing import Dict, Optional, Union, Any
from bloqade.builder.base import ParamType
from dataclasses import dataclass, field
import warnings


@dataclass
@Serializer.register
class QuEraTask(RemoteTask):
    task_id: Optional[str]
    backend: Union[QuEraBackend, MockBackend]
    task_ir: QuEraTaskSpecification
    metadata: Dict[str, ParamType]
    parallel_decoder: Optional[ParallelDecoder] = None
    task_result_ir: QuEraTaskResults = field(
        default_factory=lambda: QuEraTaskResults(
            task_status=QuEraTaskStatusCode.Unsubmitted
        )
    )

    def submit(self, force: bool = False) -> "QuEraTask":
        if not force:
            if self.task_id is not None:
                raise ValueError(
                    "the task is already submitted with %s" % (self.task_id)
                )

        self.task_id = self.backend.submit_task(self.task_ir)

        self.task_result_ir = QuEraTaskResults(task_status=QuEraTaskStatusCode.Enqueued)

        return self

    def validate(self) -> str:
        try:
            self.backend.validate_task(self.task_ir)
        except ValidationError as e:
            return str(e)

        return ""

    def fetch(self) -> "QuEraTask":
        # non-blocking, pull only when its completed
        if self.task_result_ir.task_status is QuEraTaskStatusCode.Unsubmitted:
            raise ValueError("Task ID not found.")

        if self.task_result_ir.task_status in [
            QuEraTaskStatusCode.Completed,
            QuEraTaskStatusCode.Partial,
            QuEraTaskStatusCode.Failed,
            QuEraTaskStatusCode.Unaccepted,
            QuEraTaskStatusCode.Cancelled,
        ]:
            return self

        status = self.status()
        if status in [QuEraTaskStatusCode.Completed, QuEraTaskStatusCode.Partial]:
            self.task_result_ir = self.backend.task_results(self.task_id)
        else:
            self.task_result_ir = QuEraTaskResults(task_status=status)

        return self

    def pull(self) -> "QuEraTask":
        # blocking, force pulling, even its completed
        if self.task_id is None:
            raise ValueError("Task ID not found.")

        self.task_result_ir = self.backend.task_results(self.task_id)

        return self

    def result(self) -> QuEraTaskResults:
        # blocking, caching

        if self.task_result_ir is None:
            pass
        else:
            if (
                self.task_id is not None
                and self.task_result_ir.task_status != QuEraTaskStatusCode.Completed
            ):
                self.pull()

        return self.task_result_ir

    def status(self) -> QuEraTaskStatusCode:
        if self.task_id is None:
            return QuEraTaskStatusCode.Unaccepted

        return self.backend.task_status(self.task_id)

    def cancel(self) -> None:
        if self.task_id is None:
            warnings.warn("Cannot cancel task, missing task id.")
            return

        self.backend.cancel_task(self.task_id)

    @property
    def nshots(self):
        return self.task_ir.nshots

    def _geometry(self) -> Geometry:
        return Geometry(
            sites=self.task_ir.lattice.sites,
            filling=self.task_ir.lattice.filling,
            parallel_decoder=self.parallel_decoder,
        )

    def _result_exists(self) -> bool:
        if self.task_result_ir is None:
            return False
        else:
            if self.task_result_ir.task_status in [
                QuEraTaskStatusCode.Completed,
                QuEraTaskStatusCode.Partial,
            ]:
                return True
            else:
                return False

    # def submit_no_task_id(self) -> "HardwareTaskShotResults":
    #    return HardwareTaskShotResults(hardware_task=self)


@QuEraTask.set_serializer
def _serialze(obj: QuEraTask) -> Dict[str, ParamType]:
    return {
        "task_id": obj.task_id if obj.task_id is not None else None,
        "task_ir": obj.task_ir.dict(by_alias=True, exclude_none=True),
        "metadata": obj.metadata,
        "backend": {
            f"{obj.backend.__class__.__name__}": obj.backend.dict(
                exclude=set(["access_key", "secret_key", "session_token"])
            )
        },
        "parallel_decoder": obj.parallel_decoder.dict()
        if obj.parallel_decoder
        else None,
        "task_result_ir": obj.task_result_ir.dict() if obj.task_result_ir else None,
    }


@QuEraTask.set_deserializer
def _deserializer(d: Dict[str, Any]) -> QuEraTask:
    d["task_ir"] = QuEraTaskSpecification(**d["task_ir"])
    d["task_result_ir"] = (
        QuEraTaskResults(**d["task_result_ir"]) if d["task_result_ir"] else None
    )
    d["backend"] = (
        QuEraBackend(**d["backend"]["QuEraBackend"])
        if "QuEraBackend" in d["backend"]
        else MockBackend(**d["backend"]["MockBackend"])
    )
    d["parallel_decoder"] = (
        ParallelDecoder(**d["parallel_decoder"]) if d["parallel_decoder"] else None
    )
    return QuEraTask(**d)


# class QuEraBatch(Batch, JSONInterface):
#    #futures: List[QuEraTask]
#    pass



from bloqade.visualization.ir_visualize import (
    mock_data,
    Field_wvfm,
    SpacialMod,
    assemble_field,
    assemble_pulse_panel,
    assemble_sequences,
    mock_register,
)
import itertools
from bokeh.palettes import Dark2_5
from bokeh.models import Span
from bokeh.io import save
from bokeh.layouts import row


def test_mock_data():
    shared_indicator = Span(dimension="height")

    ## Rydberg:
    dats, names, spinfo = mock_data(10)
    fig = Field_wvfm(
        colors=itertools.cycle(Dark2_5),
        data_sources=dats,
        ch_names=names,
        crx_hair_overlay=shared_indicator,
    )
    cube = SpacialMod(spinfo)
    p1 = assemble_field(cube, fig, "Detuning Fields")

    dats, names, spinfo = mock_data(10)
    fig = Field_wvfm(
        colors=itertools.cycle(Dark2_5),
        data_sources=dats,
        ch_names=names,
        crx_hair_overlay=shared_indicator,
    )
    cube = SpacialMod(spinfo)
    p2 = assemble_field(cube, fig, "Rabi amp Fields")

    dats, names, spinfo = mock_data(10)
    fig = Field_wvfm(
        colors=itertools.cycle(Dark2_5),
        data_sources=dats,
        ch_names=names,
        crx_hair_overlay=shared_indicator,
    )
    cube = SpacialMod(spinfo)
    p3 = assemble_field(cube, fig, "Rabi phase Fields")

    Panel_Pulse1 = assemble_pulse_panel([p1, p2, p3], "Rydberg")

    shared_indicator = Span(dimension="height")

    ## Hyperfine:
    dats, names, spinfo = mock_data(10)
    fig = Field_wvfm(
        colors=itertools.cycle(Dark2_5),
        data_sources=dats,
        ch_names=names,
        crx_hair_overlay=shared_indicator,
    )
    cube = SpacialMod(spinfo)
    p1 = assemble_field(cube, fig, "Detuning Fields")

    dats, names, spinfo = mock_data(10)
    fig = Field_wvfm(
        colors=itertools.cycle(Dark2_5),
        data_sources=dats,
        ch_names=names,
        crx_hair_overlay=shared_indicator,
    )
    cube = SpacialMod(spinfo)
    p2 = assemble_field(cube, fig, "Rabi amp Fields")

    dats, names, spinfo = mock_data(10)
    fig = Field_wvfm(
        colors=itertools.cycle(Dark2_5),
        data_sources=dats,
        ch_names=names,
        crx_hair_overlay=shared_indicator,
    )
    cube = SpacialMod(spinfo)
    p3 = assemble_field(cube, fig, "Rabi phase Fields")

    Panel_Pulse2 = assemble_pulse_panel([p1, p2, p3], "Hyperfine")

    Seq = assemble_sequences([Panel_Pulse1, Panel_Pulse2])

    save(row(Seq, mock_register()))



from bloqade.serialize import Serializer, dumps, loads
from beartype.typing import Union, Dict, Any
from beartype import beartype
from decimal import Decimal
from numbers import Real


LiteralType = Union[Decimal, Real]


@Serializer.register
class A:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, A):
            return self.x == __value.x and self.y == __value.y

        return False


@Serializer.register
class B:
    def __init__(self, x: float, y):
        self._x = x
        self._y = y

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, B):
            return self._x == __value._x and self._y == __value._y


@B.set_serializer
def _serialize(obj: B):
    return {"x": obj._x, "y": obj._y}


@Serializer.register
class C:
    @beartype
    def __init__(self, x: LiteralType):
        self._x = Decimal(str(x))
        self._y = 2 * x

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, C):
            return self._x == __value._x

        return False


@C.set_serializer
def _serializer(obj: C) -> Dict[str, Any]:
    return {"x": obj._x}


@C.set_deserializer
def _deserializer(d: Dict[str, Any]) -> C:
    d = {"x": Decimal(d["x"])}
    return C(**d)


def test():
    assert f'{{"{A.__module__}.A": {{"x": 1, "y": 2}}}}' == dumps(A(1, 2)), repr(
        A(1, 2)
    )
    assert f'{{"{B.__module__}.B": {{"x": 1, "y": 2}}}}' == dumps(B(1, 2)), repr(
        dumps(B(1, 2))
    )
    assert f'{{"{C.__module__}.C": {{"x": 1.23091023939020342342039402304}}}}' == dumps(
        C(Decimal("1.23091023939020342342039402304"))
    )

    assert B(C(1), A(1, 2)) == loads(dumps(B(C(1), A(1, 2))))
    assert A(1, B(1, 2)) == loads(dumps(A(1, B(1, 2))))
    assert A(A(A(A(1, 2), 3), 4), C(4)) == loads(dumps(A(A(A(A(1, 2), 3), 4), C(4))))

    a = A("a", "b")
    a = A(a, "c")
    a = A(1, a)
    a = A(B(a, C(1)), B("A", a))
    a = A(B(1, 2), A(a, C(2)))

    assert a == loads(dumps(a))


test()



from dataclasses import dataclass
from numpy.typing import NDArray
import numpy as np
from enum import Enum

from bloqade.serialize import Serializer


@dataclass(frozen=True)
class AtomType:
    def string_to_integer(self, fock_state: str) -> int:
        state_string = fock_state.replace("|", "").replace(">", "")

        state_int = 0
        shift = 1
        for local_state in state_string:
            state_int += shift * self.str_to_int[local_state]
            shift *= self.n_level

        return state_int

    def integer_to_string(self, state_int: int, n_atoms: int) -> str:
        state_string = ""
        local_state = state_int % self.n_level

        state_string = state_string + self.int_to_str[local_state]
        state_int //= self.n_level

        while state_int > 0:
            local_state = state_int % self.n_level

            state_string = state_string + self.int_to_str[local_state]
            state_int //= self.n_level

        if len(state_string) < n_atoms:
            state_string = state_string + "g" * (n_atoms - len(state_string))

        return f"|{state_string}>"

    def is_rydberg_at(self, configurations: NDArray, index: int) -> NDArray:
        return self.is_state_at(configurations, index, self.State.Rydberg)

    def swap_state_at(
        self, configurations: NDArray, index: int, state_1: int, state_2: int
    ) -> NDArray:
        raise NotImplementedError

    def transition_state_at(
        self, configurations: NDArray, index: int, fro: int, to: int
    ) -> NDArray:
        raise NotImplementedError

    def is_state_at(self, configurations: NDArray, index: int, state: int) -> NDArray:
        raise NotImplementedError

    def __hash__(self):
        return hash(self.__class__)


@dataclass(frozen=True)
@Serializer.register
class ThreeLevelAtomType(AtomType):
    n_level = 3
    str_to_int = {"g": 0, "h": 1, "r": 2}
    int_to_str = ["g", "h", "r"]

    class State(int, Enum):
        Ground = 0
        Hyperfine = 1
        Rydberg = 2

    def is_state_at(cls, configurations: NDArray, index: int, state):
        if not isinstance(state, cls.State):
            raise ValueError(f"state: {state} is not a valid state for {cls.__name__}.")

        mask = (configurations // 3**index) % 3
        return mask == state.value

    def swap_state_at(
        self, configurations: NDArray, index: int, state_1: State, state_2: State
    ) -> NDArray:
        state_1 = self.State(state_1)
        state_2 = self.State(state_2)

        output = configurations.copy()

        mask_1 = self.is_state_at(configurations, index, state_1)
        mask_2 = self.is_state_at(configurations, index, state_2)
        delta = state_2.value - state_1.value

        if delta < 0:
            mask_1, mask_2 = mask_2, mask_1
            delta = -delta

        shift_value = np.array(delta * 3**index, dtype=configurations.dtype)

        output[mask_1] += shift_value
        output[mask_2] -= shift_value

        np.logical_or(mask_1, mask_2, out=mask_1)

        return mask_1, output[mask_1]

    def transition_state_at(
        self, configurations: NDArray, index: int, fro: State, to: State
    ) -> NDArray:
        fro = self.State(fro)
        to = self.State(to)

        input_configs = self.is_state_at(configurations, index, fro)
        output_configs = configurations[input_configs]

        delta = to.value - fro.value

        if delta < 0:
            delta = -delta
            output_configs -= np.array(delta * 3**index, dtype=output_configs.dtype)
        else:
            output_configs += np.array(delta * 3**index, dtype=output_configs.dtype)

        return (input_configs, output_configs)


@dataclass(frozen=True)
@Serializer.register
class TwoLevelAtomType(AtomType):
    n_level = 2
    str_to_int = {"g": 0, "r": 1}
    int_to_str = ["g", "r"]

    class State(int, Enum):
        Ground = 0
        Rydberg = 1

    def is_state_at(self, configurations: NDArray, index: int, state: State):
        state = self.State(state)

        mask = ((configurations >> index) & 1) == 1

        if state == self.State.Ground:
            return np.logical_not(mask)

        return mask

    def swap_state_at(
        self, configurations: NDArray, index: int, state_1: State, state_2: State
    ) -> NDArray:
        state_1 = self.State(state_1)
        state_2 = self.State(state_2)
        return slice(None), configurations ^ (1 << index)

    def transition_state_at(
        self, configurations: NDArray, index: int, fro: State, to: State
    ) -> NDArray:
        fro = self.State(fro)
        to = self.State(to)

        input_configs = self.is_state_at(configurations, index, fro)
        output_configs = configurations[input_configs]
        return (input_configs, output_configs ^ (1 << index))


ThreeLevelAtom = ThreeLevelAtomType()
TwoLevelAtom = TwoLevelAtomType()



from bloqade.serialize import Serializer
from bloqade.task.base import Geometry, LocalTask
from bloqade.emulate.ir.emulator import EmulatorProgram
from bloqade.emulate.codegen.hamiltonian import (
    RydbergHamiltonianCodeGen,
    CompileCache,
)
from bloqade.emulate.ir.state_vector import AnalogGate

from bloqade.submission.ir.task_results import (
    QuEraTaskResults,
    QuEraShotResult,
    QuEraTaskStatusCode,
    QuEraShotStatusCode,
)
from beartype.typing import Dict, Any
from bloqade.builder.base import ParamType
from dataclasses import dataclass
from typing import Optional


@dataclass
@Serializer.register
class BloqadeTask(LocalTask):
    shots: int
    emulator_ir: EmulatorProgram
    metadata: Dict[str, ParamType]
    compile_cache: Optional[CompileCache] = None
    task_result_ir: Optional[QuEraTaskResults] = None

    def _geometry(self) -> Geometry:
        sites = [tuple(map(float, site)) for site in self.emulator_ir.register.sites]

        return Geometry(sites=sites, filling=[1 for _ in sites], parallel_decoder=None)

    def result(self) -> QuEraTaskResults:
        return self.task_result_ir

    def run(
        self,
        solver_name: str = "dop853",
        atol: float = 1e-14,
        rtol: float = 1e-7,
        nsteps: int = 2_147_483_647,
        interaction_picture: bool = False,
    ) -> "BloqadeTask":
        hamiltonian = RydbergHamiltonianCodeGen(self.compile_cache).emit(
            self.emulator_ir
        )
        options = dict(
            solver_name=solver_name,
            atol=atol,
            rtol=rtol,
            nsteps=nsteps,
            interaction_picture=interaction_picture,
        )
        shots_array = AnalogGate(hamiltonian).run(
            self.shots, project_hyperfine=True, **options
        )

        shot_outputs = []
        for shot in shots_array[:]:
            shot_result = QuEraShotResult(
                shot_status=QuEraShotStatusCode.Completed,
                # TODO: make the pre_sesquence and post_sequence match the
                # empty sites in the original definition of the register
                pre_sequence=[1 for _ in shot],
                # flip the bits so that 1 = ground state and 0 = excited state
                post_sequence=list(1 - shot),
            )
            shot_outputs.append(shot_result)

        self.task_result_ir = QuEraTaskResults(
            task_status=QuEraTaskStatusCode.Completed, shot_outputs=shot_outputs
        )

        return self


@BloqadeTask.set_serializer
def _serialze(obj: BloqadeTask) -> Dict[str, Any]:
    return {
        "shots": obj.shots,
        "emulator_ir": obj.emulator_ir,
        "metadata": obj.metadata,
        "task_result_ir": obj.task_result_ir.dict() if obj.task_result_ir else None,
    }


@BloqadeTask.set_deserializer
def _deserialize(d: Dict[str, Any]) -> BloqadeTask:
    d["task_result_ir"] = (
        QuEraTaskResults(**d["task_result_ir"]) if d["task_result_ir"] else None
    )
    return BloqadeTask(**d)



# from numbers import Real
from bloqade.visualization import display_ir
from bloqade.ir.control.sequence import SequenceExpr
from bloqade.ir.location.base import AtomArrangement, ParallelRegister
from bloqade.ir.tree_print import Printer
from beartype.typing import Union
from pydantic.dataclasses import dataclass


# NOTE: this is just a dummy type bundle geometry and sequence
#       information together and forward them to backends.
@dataclass(frozen=True)
class AnalogCircuit:
    """AnalogCircuit is a dummy type that bundle register and sequence together."""

    atom_arrangement: Union[ParallelRegister, AtomArrangement]
    sequence: SequenceExpr

    @property
    def register(self):
        """Get the register of the program.

        Returns:
            register (Union["AtomArrangement", "ParallelRegister"])

        Note:
            If the program is built with
            [`parallelize()`][bloqade.builder.emit.Emit.parallelize],
            The the register will be a
            [`ParallelRegister`][bloqade.ir.location.base.ParallelRegister].
            Otherwise it will be a
            [`AtomArrangement`][bloqade.ir.location.base.AtomArrangement].
        """
        return self.atom_arrangement

    def __eq__(self, other):
        if isinstance(other, AnalogCircuit):
            return (self.register == other.register) and (
                self.sequence == other.sequence
            )

        return False

    def __str__(self):
        out = ""
        if self.register is not None:
            out += self.register.__str__()

        out += "\n"

        if self.sequence is not None:
            out += self.sequence.__str__()

        return out

    def print_node(self):
        return "AnalogCircuit"

    def children(self):
        return {"register": self.atom_arrangement, "sequence": self.sequence}

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def figure(self, **assignments):
        fig_reg = self.register.figure(**assignments)
        fig_seq = self.sequence.figure(**assignments)
        return fig_seq, fig_reg

    def show(self, **assignments):
        """Interactive visualization of the program

        Args:
            **assignments: assigning the instance value (literal) to the
                existing variables in the program

        """
        display_ir(self, assignments)



from pydantic.dataclasses import dataclass
import bloqade.ir.analog_circuit as analog_circuit
import bloqade.ir.location as location
import bloqade.ir.scalar as scalar
import bloqade.ir.control.waveform as waveform
import bloqade.ir.control.sequence as sequence
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.field as field

from bloqade.ir.visitor.analog_circuit import AnalogCircuitVisitor
from bloqade.ir.visitor.location import LocationVisitor
from bloqade.ir.visitor.waveform import WaveformVisitor
from bloqade.ir.visitor.scalar import ScalarVisitor
from beartype.typing import Set, Any, FrozenSet


@dataclass(frozen=True)
class ScanVariableResults:
    scalar_vars: FrozenSet[str]
    vector_vars: FrozenSet[str]


class ScanVariablesScalar(ScalarVisitor):
    def __init__(self):
        self.vars = set()

    def visit_literal(self, ast: scalar.Literal) -> Any:
        pass

    def visit_assigned_variable(self, ast: scalar.AssignedVariable) -> Any:
        self.vars.add(ast.name)

    def visit_variable(self, ast: scalar.Variable) -> Any:
        self.vars.add(ast.name)

    def visit_add(self, ast: scalar.Add) -> Any:
        self.visit(ast.lhs)
        self.visit(ast.rhs)

    def visit_mul(self, ast: scalar.Mul) -> Any:
        self.visit(ast.lhs)
        self.visit(ast.rhs)

    def visit_div(self, ast: scalar.Div) -> Any:
        self.visit(ast.lhs)
        self.visit(ast.rhs)

    def visit_interval(self, ast: scalar.Interval) -> Any:
        if ast.start is not None:
            self.visit(ast.start)

        if ast.stop is not None:
            self.visit(ast.stop)

    def visit_slice(self, ast: scalar.Slice) -> Any:
        self.visit(ast.expr)
        self.visit(ast.interval)

    def visit_max(self, ast: scalar.Max) -> Any:
        list(map(self.visit, ast.exprs))

    def visit_min(self, ast: scalar.Min) -> Any:
        list(map(self.visit, ast.exprs))

    def visit_negative(self, ast: scalar.Negative) -> Any:
        self.visit(ast.expr)

    def emit(self, ast: scalar.Scalar) -> Set[str]:
        self.visit(ast)
        return self.vars


class ScanVariablesWaveform(WaveformVisitor):
    def __init__(self):
        self.vars = set()
        self.scalar_visitor = ScanVariablesScalar()

    def visit_constant(self, ast: waveform.Constant) -> Any:
        self.vars = self.vars.union(self.scalar_visitor.emit(ast.value))
        self.vars = self.vars.union(self.scalar_visitor.emit(ast.duration))

    def visit_linear(self, ast: waveform.Linear) -> Any:
        self.vars = self.vars.union(self.scalar_visitor.emit(ast.start))
        self.vars = self.vars.union(self.scalar_visitor.emit(ast.stop))
        self.vars = self.vars.union(self.scalar_visitor.emit(ast.duration))

    def visit_poly(self, ast: waveform.Poly) -> Any:
        for coeff in ast.coeffs:
            self.vars = self.vars.union(self.scalar_visitor.emit(coeff))
        self.vars = self.vars.union(self.scalar_visitor.emit(ast.duration))

    def visit_python_fn(self, ast: waveform.PythonFn) -> Any:
        self.vars = self.vars.union(self.scalar_visitor.emit(ast.duration))
        for parameter in ast.parameters:
            self.vars = self.vars.union(self.scalar_visitor.emit(parameter))

    def visit_negative(self, ast: waveform.Negative) -> Any:
        self.visit(ast.waveform)

    def visit_record(self, ast: waveform.Record):
        self.visit(ast.waveform)
        self.vars.add(ast.var.name)

    def visit_add(self, ast: waveform.Add) -> Any:
        self.visit(ast.left)
        self.visit(ast.right)

    def visit_scale(self, ast: waveform.Scale) -> Any:
        self.visit(ast.waveform)
        self.vars = self.vars.union(self.scalar_visitor.emit(ast.scalar))

    def visit_alligned(self, ast: waveform.AlignedWaveform) -> Any:
        self.visit(ast.waveform)
        if isinstance(ast.value, scalar.Scalar):
            self.vars = self.vars.union(self.scalar_visitor.emit(ast.value))

    def visit_sample(self, ast: waveform.Sample) -> Any:
        self.vars = self.vars.union(self.scalar_visitor.emit(ast.dt))
        self.visit(ast.waveform)

    def visit_append(self, ast: waveform.Append) -> Any:
        list(map(self.visit, ast.waveforms))

    def visit_slice(self, ast: waveform.Slice) -> Any:
        self.vars = self.vars.union(self.scalar_visitor.emit(ast.interval))
        self.visit(ast.waveform)

    def visit_smooth(self, ast: waveform.Smooth) -> Any:
        self.visit(ast.waveform)
        self.vars = self.vars.union(self.scalar_visitor.emit(ast.radius))

    def emit(self, ast: waveform.Waveform) -> Set[str]:
        self.visit(ast)
        return self.vars


class ScanVariablesLocation(LocationVisitor):
    def __init__(self):
        self.vars = set()
        self.scalar_visitor = ScanVariablesScalar()

    def visit_bravais(self, ast: location.BoundedBravais) -> Any:
        self.vars = self.vars.union(self.scalar_visitor.emit(ast.lattice_spacing))

    def visit_chain(self, ast: location.Chain) -> Any:
        self.visit_bravais(ast)

    def visit_honeycomb(self, ast: location.Honeycomb) -> Any:
        self.visit_bravais(ast)

    def visit_kagome(self, ast: location.Kagome) -> Any:
        self.visit_bravais(ast)

    def visit_lieb(self, ast: location.Lieb) -> Any:
        self.visit_bravais(ast)

    def visit_square(self, ast: location.Square) -> Any:
        self.visit_bravais(ast)

    def visit_triangular(self, ast: location.Triangular) -> Any:
        self.visit_bravais(ast)

    def visit_rectangular(self, ast: location.Rectangular) -> Any:
        self.scalar_visitor.emit(ast.lattice_spacing_x)
        self.scalar_visitor.emit(ast.lattice_spacing_y)

    def visit_list_of_locations(self, ast: location.ListOfLocations) -> Any:
        for loc in ast.location_list:
            self.visit(loc)

    def visit_location_info(self, ast: location.LocationInfo) -> Any:
        for coord in ast.position:
            self.vars = self.vars.union(self.scalar_visitor.emit(coord))

    def visit_parallel_register(self, ast: location.ParallelRegister) -> Any:
        self.visit(ast._register)
        self.vars = self.vars.union(self.scalar_visitor.emit(ast._cluster_spacing))

    def emit(self, ast: location.AtomArrangement) -> Set[str]:
        self.visit(ast)
        return self.vars


class ScanVariablesAnalogCircuit(AnalogCircuitVisitor):
    def __init__(self):
        self.scalar_vars = set()
        self.vector_vars = set()
        self.waveform_visitor = ScanVariablesWaveform()
        self.scalar_visitor = ScanVariablesScalar()
        self.location_visitor = ScanVariablesLocation()

    def visit_uniform_modulation(self, ast: field.UniformModulation) -> Any:
        pass

    def visit_scaled_locations(self, ast: field.ScaledLocations) -> Any:
        for value in ast.value.values():
            self.scalar_vars = self.scalar_vars.union(self.scalar_visitor.emit(value))

    def visit_run_time_vector(self, ast: field.RunTimeVector) -> Any:
        self.vector_vars.add(ast.name)

    def visit_assigned_run_time_vector(self, ast: field.AssignedRunTimeVector) -> Any:
        self.vector_vars.add(ast.name)

    def visit_field(self, ast: field.Field) -> Any:
        list(map(self.visit, ast.drives.keys()))
        for value in ast.drives.values():
            self.visit(value)

    def visit_pulse(self, ast: pulse.Pulse) -> Any:
        list(map(self.visit, ast.fields.values()))

    def visit_append_pulse(self, ast: pulse.Append) -> Any:
        list(map(self.visit, ast.pulses))

    def visit_slice_pulse(self, ast: pulse.Slice) -> Any:
        self.scalar_vars = self.scalar_vars.union(
            self.scalar_visitor.emit(ast.interval)
        )
        self.visit(ast.pulse)

    def visit_named_pulse(self, ast: pulse.NamedPulse) -> Any:
        self.visit(ast.pulse)

    def visit_sequence(self, ast: sequence.Sequence) -> Any:
        list(map(self.visit, ast.pulses.values()))

    def visit_append_sequence(self, ast: sequence.Append) -> Any:
        list(map(self.visit, ast.sequences))

    def visit_slice_sequence(self, ast: sequence.Slice) -> Any:
        self.scalar_vars = self.scalar_vars.union(
            self.scalar_visitor.emit(ast.interval)
        )
        self.visit(ast.sequence)

    def visit_named_sequence(self, ast: sequence.NamedSequence) -> Any:
        self.visit(ast.sequence)

    def visit_waveform(self, ast: waveform.Waveform) -> Any:
        self.scalar_vars = self.scalar_vars.union(self.waveform_visitor.emit(ast))

    def visit_register(self, ast: location.AtomArrangement) -> Any:
        self.scalar_vars = self.scalar_vars.union(self.location_visitor.emit(ast))

    def visit_parallel_register(self, ast: location.ParallelRegister) -> Any:
        self.scalar_vars = self.scalar_vars.union(self.location_visitor.emit(ast))

    def visit_analog_circuit(self, ast: analog_circuit.AnalogCircuit) -> Any:
        self.visit(ast.register)
        self.visit(ast.sequence)

    def emit(self, ast: analog_circuit.AnalogCircuit) -> ScanVariableResults:
        self.visit(ast)
        return ScanVariableResults(self.scalar_vars, self.vector_vars)



from itertools import repeat
from beartype.typing import Optional, List, Dict, Set, Sequence, Union
from bloqade.builder.typing import ParamType
from bloqade.builder.base import Builder
from bloqade.builder.pragmas import Parallelizable, AddArgs, BatchAssignable
from bloqade.builder.backend import BackendRoute
from numbers import Real
from decimal import Decimal
import numpy as np


class CastParams:
    def __init__(self, n_sites: int, scalar_vars: Set[str], vector_vars: Set[str]):
        self.n_sites = n_sites
        self.scalar_vars = scalar_vars
        self.vector_vars = vector_vars

    def cast_scalar_param(self, value: ParamType, name: str) -> Decimal:
        if isinstance(value, (Real, Decimal)):
            return Decimal(str(value))

        raise TypeError(
            f"assign parameter '{name}' must be a real number, "
            f"found type: {type(value)}"
        )

    def cast_vector_param(
        self,
        value: Union[np.ndarray, List[ParamType]],
        name: str,
    ) -> List[Decimal]:
        if isinstance(value, np.ndarray):
            value = value.tolist()

        if isinstance(value, (list, tuple)):
            if len(value) != self.n_sites:
                raise ValueError(
                    f"assign parameter '{name}' must be a list of length "
                    f"{self.n_sites}, found length: {len(value)}"
                )
            return list(map(self.cast_scalar_param, value, repeat(name, len(value))))

        raise TypeError(
            f"assign parameter '{name}' must be a list of real numbers, "
            f"found type: {type(value)}"
        )

    def cast_params(self, params: Dict[str, ParamType]) -> Dict[str, ParamType]:
        checked_params = {}

        for name, value in params.items():
            if name not in self.scalar_vars and name not in self.vector_vars:
                raise ValueError(
                    f"assign parameter '{name}' is not found in analog circuit."
                )
            if name in self.vector_vars:
                checked_params[name] = self.cast_vector_param(value, name)
            else:
                checked_params[name] = self.cast_scalar_param(value, name)

        return checked_params


class AssignBase(Builder):
    pass


class Assign(BatchAssignable, AddArgs, Parallelizable, BackendRoute, AssignBase):
    def __init__(
        self, assignments: Dict[str, ParamType], parent: Optional[Builder] = None
    ) -> None:
        from bloqade.ir.analysis.scan_variables import ScanVariablesAnalogCircuit

        super().__init__(parent)

        circuit = self.parse_circuit()
        variables = ScanVariablesAnalogCircuit().emit(circuit)

        self._static_params = CastParams(
            circuit.register.n_sites, variables.scalar_vars, variables.vector_vars
        ).cast_params(assignments)


class BatchAssign(AddArgs, Parallelizable, BackendRoute, AssignBase):
    def __init__(
        self, assignments: Dict[str, List[ParamType]], parent: Optional[Builder] = None
    ) -> None:
        from bloqade.ir.analysis.scan_variables import ScanVariablesAnalogCircuit

        super().__init__(parent)

        if len(assignments) == 0:
            self._batch_params = []
            return

        circuit = self.parse_circuit()
        variables = ScanVariablesAnalogCircuit().emit(circuit)

        if not len(np.unique(list(map(len, assignments.values())))) == 1:
            raise ValueError(
                "all the assignment variables need to have same number of elements."
            )

        tuple_iterators = [
            zip(repeat(name), values) for name, values in assignments.items()
        ]

        caster = CastParams(
            circuit.register.n_sites, variables.scalar_vars, variables.vector_vars
        )

        self._batch_params = list(
            map(caster.cast_params, map(dict, zip(*tuple_iterators)))
        )


class ListAssign(AddArgs, Parallelizable, BackendRoute, AssignBase):
    def __init__(
        self,
        batch_params: Sequence[Dict[str, ParamType]],
        parent: Optional[Builder] = None,
    ) -> None:
        from bloqade.ir.analysis.scan_variables import ScanVariablesAnalogCircuit

        super().__init__(parent)

        circuit = self.parse_circuit()
        variables = ScanVariablesAnalogCircuit().emit(circuit)
        caster = CastParams(
            circuit.register.n_sites, variables.scalar_vars, variables.vector_vars
        )

        keys = set([])
        for params in batch_params:
            keys.update(params.keys())

        for batch_num, params in enumerate(batch_params):
            curr_keys = set(params.keys())
            missing_keys = keys.difference(curr_keys)
            if missing_keys:
                raise ValueError(
                    f"Batch {batch_num} missing key(s): {tuple(missing_keys)}."
                )

        self._batch_params = list(map(caster.cast_params, batch_params))
